"""#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: fmaschietto, bcallen95"""

import mdigest.core.auxiliary   as aux
import mdigest.core.toolkit     as tk
from   mdigest.core.imports     import *
from   mdigest.utils.pdbhandler import *
from   pymol                    import cmd
from   sklearn.preprocessing    import MinMaxScaler


class ProcCorr:
    """Process correlation matrix to make the desired arrays amenable for visualization."""
    def __init__(self):
        """
        Description
        -----------
        Process correlation matrix to make the desired arrays amenable for visualization.
        Possible actions to apply on desired correlation matrices include filtering, pruning upon distance matrix,
        dumping short-range entries, dumping long-range entries.

        Attributes
        ----------
        self.mda_u: mda.Universe object,
            mda.Universe
        self.matrix_dict: dict,
            dictionary containing all matrix entries for which to produce network
        self.pruning_distmat: False or str,
            - False: no pruning distance matrix provided
            - str: use name of distance matrix in matrix_dict
        self.atom_group_selection:  str,
            typically 'name CA', should match selection used to generate the correlation matrix
        self.lower_thr: float,
            A distance value in Angstrom unit. For example, it is good to remove
            high correlations for residues within less than 5.0 Angstrom distance
            to have a clear visualization. Default value np.min(self.distance).
        self.upper_thr: float,
            A distance value in Angstrom unit. The residues with this value or higher
            will not be visualized with PyMol or VMD. Default value is 9999.0 Angstrom.

        self.loc_factor: float,
            locality factor - applied as distance_matrix * 1/loc_factor
        self.inv_loc_factor,
            locality factor - applied as loc_factor * 1/distance_matrix
        self.df: pd.DataFrame,
            ouptut dataframe

        """
        self.mda_u = None
        self.matrix_dict = {}
        self.pruning_distmat = None
        self.atom_group_selstr = None
        self.atom_group_selection = None
        self.lower_thr = None
        self.upper_thr = None
        self.loc_factor = None
        self.inv_loc_factor = None
        self.df = None
        #### output ####

        self.outname = None
        self.outpath = None


    def source_universe(self, universe):
        """
        Source mda.Universe

        Parameters
        ----------
        universe: mda.Universe object

        """
        self.mda_u = universe


    def get_selection_fromMDS(self, MDS):
        """
        Retrieve atomstring selection

        Parameters
        ----------
        MDS: mdigest.MDS object


        """
        self.atom_group_selstr = MDS.atom_group_selstr


    def set_selection(self, atom_group_selstr):
        """
        Set atom group selection string

        Parameters
        ----------
        atom_group_selstr: str
            selection string
                example: 'name CA'
        """
        self.atom_group_selstr = atom_group_selstr
        print('@>: atom group selection string set to {}'.format(self.atom_group_selstr))


    def load_matrix_dictionary(self, matrix_dictionary):
        """
        Populate matrix_dictionary attribute with kwargs

        Parameters
        ----------
        matrix_dictionary: dict,
            dictionary with format of {'matrix_label': np.ndarray} containing
            correlation matrices to visualize
        """

        self.matrix_dict = OrderedDict(matrix_dictionary)


    def populate_attributes(self, matrixdict):
        """
        Create attributes of ProcCorr class corresponding to the keys of matrix_dictionary

        Parameters
        ----------
        matrixdict: dict,
            example: matrixdict = {'matrix_label': np.ndarray}
        """

        df = pd.DataFrame(columns=['Source', 'Target'])
        for matrixtype, matrix in matrixdict.items():
            print('@> setting {} attribute'.format(matrixtype))
            setattr(self, matrixtype, matrix)


    def set_thresholds(self, unit='au', prune_upon=False, **kwargs):
        """

        Parameters
        ----------
        unit: str,
            unit, default a.u., possible values 'nm', 'au'
        prune_upon: False or str,
            where str is the matrix_label of the array to use for pruning the correlations
        kwargs: dict,
            - lower_thr
            - upper_thr
            - loc_factor
            - inv_loc_factor
        """

        if unit == 'nm':
            factor = 10
        else:
            factor = 1

        if prune_upon is not False:
            if isinstance(prune_upon, str):
                if hasattr(self, prune_upon):
                    setattr(self, 'pruning_distmat', getattr(self, str(prune_upon)))
                else:
                    print('Warning: object has no attribute {}\nset attributes by calling retrieve_matrix() ')
            elif isinstance(prune_upon, np.ndarray):
                self.pruning_distmat = prune_upon

            if kwargs.__contains__('lower_thr'):
                self.lower_thr = kwargs['lower_thr']
            else:
                self.lower_thr = np.min(self.pruning_distmat * factor)
            print('@>: lower distance threshold is set to {}'.format(self.lower_thr))
            if kwargs.__contains__('upper_thr'):
                self.upper_thr = kwargs['upper_thr']
            else:
                self.upper_thr = np.max(self.pruning_distmat * factor)
            print('@>: upper distance threshold is set to {}'.format(self.upper_thr))

            if kwargs.__contains__('loc_factor'):
                self.loc_factor = kwargs['loc_factor']
            print('@>: loc_factor is set to {}'.format(self.loc_factor))
            if kwargs.__contains__('inv_loc_factor'):
                self.inv_loc_factor = kwargs['inv_loc_factor']
            print('@>: inv_loc_factor is set to {}'.format(self.inv_loc_factor))

        else:
            print('@>: No distance matrix has been loaded into class')
            print('@>: Load distance matrix using `retrieve_matrix(mat, matrixtype="distance")`')
            if kwargs.__contains__('lower_thr'):
                self.lower_thr = kwargs['lower_thr']
            else:
                self.lower_thr = np.min(self.pruning_distmat * factor)
            print('@>: lower distance threshold is set to {}'.format(self.lower_thr))
            if kwargs.__contains__('upper_thr'):
                self.upper_thr = kwargs['upper_thr']
            else:
                self.upper_thr = np.max(self.pruning_distmat * factor)
            print('@>: upper distance threshold is set to {}'.format(self.upper_thr))


    def filter_by_distance(self, matrixtype, distmat=False):
        """
        Zero out correlations lower than self.lower_str and higher than self.upper_str in a given
        correlation matrix. If residues are closer to each other than a certain distance
        (self.lower_thr), make these correlations zero. If residues are farther to
        each other than a certain distance (self.upper_thr), make these correlations
        also zero. This filter can be applied to select correlations falling within
        distance window of interest.

        Parameters
        ----------
        matrixtype: str,
            used to select a desired correlation matrix and also used as prefix for the output files.

        distmat: bool,
            default is False, which results in pruning based on correlation values.
            Upper and lower thresholds for pruning are set by call to `set_thresholds()`
        """

        matrix = getattr(self, matrixtype)
        dist_mat = None

        if distmat:
            dist_mat = self.pruning_distmat
            if self.loc_factor is None and self.inv_loc_factor is None:
                print('Pruning based on distance matrix values')

                if self.upper_thr is None or self.lower_thr is None:
                    self.upper_thr = 5.
                    self.lower_thr = 0.
                    print('@>: setting upper_thr and lower_thr for matrix filtering to default values')
                    print('@>: upper_thr = {}'.format(self.upper_thr))
                    print('@>: lower_thr = {}'.format(self.lower_thr))
                    self.set_thresholds()
                print(
                    f'@>: filtering correlations corresponding to inter-residue distances lower than {self.lower_thr} Angstrom and')
                print(f'@>: higher than {self.upper_thr} Angstrom')

        elif distmat == True and self.pruning_distmat is None:
            print('@>: call `set_thresholds(prune_upon=nd.array(distance_mat))`, where nd.array(distance_mat) is the'
                  'upon which to prune the selected correlation / adjancency matrices ')
            exit()

        else:
            print('Pruning based on correlation values')
            if self.upper_thr is None or self.lower_thr is None:
                self.upper_thr = 1.
                self.lower_thr = 0.5
                print('@>: setting upper_thr and lower_thr for matrix filtering to default values')
                print('@>: upper_thr = {}'.format(self.upper_thr))
                print('@>: lower_thr = {}'.format(self.lower_thr))

            print(f'@>: filtering correlations lower than {self.lower_thr}  and')
            print(f'@>: higher than {self.upper_thr}  inter-residue distances')

        if distmat:
            if self.loc_factor is None and self.inv_loc_factor is None:
                for i in range(0, len(matrix)):
                    for j in range(i + 1, len(matrix)):
                        if ((dist_mat[i, j] < self.lower_thr) or
                                (dist_mat[i, j] >= self.upper_thr)):
                            matrix[i, j] = 0.0
                            matrix[j][i] = 0.0
            else:
                if self.loc_factor is not None:
                    matrix = aux.filter_adjacency(matrix, distmat=dist_mat, loc_factor=self.loc_factor)

                elif self.inv_loc_factor is not None:
                    matrix = aux.filter_adjacency_inv(matrix, distmat=dist_mat, loc_factor=self.inv_loc_factor)

        else:
            for i in range(0, len(matrix)):
                for j in range(i + 1, len(matrix)):
                    if ((matrix[i, j] < self.lower_thr) or
                            (matrix[i, j] >= self.upper_thr)):
                        matrix[i, j] = 0.0
                        matrix[j][i] = 0.0

        self.matrix_dict[matrixtype] = matrix
        super().__setattr__(matrixtype, matrix)


    def to_df(self, normalize=False, **kwargs):
        """
        Save filtered matrices to pandas DataFrame

        Parameters
        ----------
        normalize: bool,
        kwargs: dict,
            - which: str,
                name of matrix (column) on which to apply normalization
            - to_range: range,
                range for normalization
        """

        df = pd.DataFrame(columns=['Source', 'Target'])

        for matrixtype, matrix in self.matrix_dict.items():
            tmp_df = pd.DataFrame(matrix)
            tmp_df = tmp_df.rename_axis('Source') \
                .reset_index() \
                .melt('Source', value_name=matrixtype, var_name='Target') \
                .query('Source != Target') \
                .reset_index(drop=True)

            if df.empty:
                df = pd.merge(left=df, right=tmp_df, how='outer')
            else:
                df = pd.merge(left=df, right=tmp_df)
            if normalize:
                if 'which' in kwargs.keys() and matrixtype in kwargs['which']:
                    print('@>: apply normalization on {} column'.format(matrixtype))

                    if 'to_range' in kwargs.keys():
                        print('@>: map {} column to {} range'.format(matrixtype, str(kwargs['to_range'])))
                        df[matrixtype] = MinMaxScaler(kwargs['to_range']).fit_transform(
                            np.array(df[matrixtype]).reshape(-1, 1))

                elif 'which' in kwargs.keys() and matrixtype not in kwargs['which']:
                    print('@>: skip normalization on {} column'.format(matrixtype))

                else:
                    print('@>: apply normalization on {} column'.format(matrixtype))
                    print('@>: map {} column to {} range'.format(matrixtype, str('(0, 1)')))
                    df[matrixtype] = MinMaxScaler().fit_transform(
                        np.array(df[matrixtype]).reshape(-1, 1))

        self.df = df

    def set_outputparams(self, params):
        """
        Set ouptut parameters
        """
        if params.__contains__('outpath'):
            self.outpath = params['outpath']
            print(self.outpath)
        else:
            self.outpath = './'

    def select_frame_coordinates(self, frame):
        """
        Select frame
        """
        ids = np.asarray(self.atom_group_selection.atoms.indices)
        return self.mda_u.trajectory[frame].positions[ids, :]

    def writePDBforframe(self, frame, outpdb):
        """
        Write PDB for a selected frame

        Parameters
        ----------
        frame: int,
            selected frame for which to write PDB

        outpdb: str,
            output filename
        """
        topology_df = mdatopology_to_dataframe(self.mda_u, frame=frame)
        df_to_pdb(topology_df, outpdb)

    def corrNetworkPymol(self, pdb_file, corr_matrix, pml_out_file, frame=0, selection=None,
                         lthr_filter=None, uthr_filter=None, edge_scaling=1, chainblocks=True,
                         **kwargs):
        """
        A useful practice is to inspect the shape of the correlation networks.
        This function provides a way to visualize the correlation patterns on the protein
        structure. The function saves a pml file that contains the user selected
        correlation values which can be executed in pymol to produce a png of the correlation
        ``[***]``

        Parameters
        ----------
        pdb_file: str,
            filename (path+filename+extension) of the PDB that will then be read in pymol. If the file is not found, the utils module is
            called to write a pdb file at the path specified by pdb_file. It is crucial that this pdb file corresponds to the trajectory frame used for reading the coordinates
            (specified by the variable ``frame``)
        frame: int,
            frame number
        corr_matrix: np.ndarray square matrix of floats,
            correlation matrix
        pml_out_file: str,
            output pml rootname
        selection: None or MDAnalysis AtomGroup object,
            - if None: selection is overwritten with self.atom_group_selstr
            - else takes in MDAnalysis AtomGroup.
        lthr_filter: float,
            lthr_filter and uthr_filter can be used to visualize correlations within an interval
            use lthr_filter to filter out correlation values below this value. Only correlation values greater than lthr_filter will be written to PML file for visualization
        uthr_filter: float,
            lthr_filter and uthr_filter can be used to visualize correlations within an interval
            use uthr_filter to filter out correlation values above this value. Only correlation values equal or lower than uthr_filter will be written to PML file for visualization
        edge_scaling: float,
            adjust radius of cylinders to be displayed in pymol.
        edge_scaling: float,
            multiplicative factor which can be used to scale the correlation values. Recommended values are between 0.01-2.00.
        chainblocks: bool,
            If True and universe contains multiple chains, separate file for inter and intra-chain correlations are printed out.
        **kwargs: dict
            provide a 'cmap_name'='name of colormap' to change the coloring scheme for edges. default is 'coolwarm'

        See Also
        --------
        ``[***]`` function adapted from https://github.com/tekpinar/correlationplus/blob/master/correlationplus/
        """

        if tk.file_exists(pdb_file):
            print('%s file will be used in pymol ' % pdb_file)
            pass
        else:
            print('write PDB for frame %d' % frame)
            self.writePDBforframe(frame, pdb_file)

        try:
            chainIDs = self.mda_u.select_atoms('all').atoms.chainIDs
        except ValueError:
            try:
                chainIDs = kwargs['chainIDs']
            except KeyError:
                chainIDs = np.asarray(['A'] * len(self.mda_u.select_atoms('all').atoms))
                print(chainIDs)
            self.mda_u.add_TopologyAttr('chainIDs', chainIDs)
            print(self.mda_u.select_atoms('all').atoms.chainIDs)

        if selection is None:
            self.atom_group_selection = self.mda_u.select_atoms(self.atom_group_selstr)
        else:
            self.atom_group_selection = self.mda_u.select_atoms(selection)

        ags = self.atom_group_selection

        if lthr_filter is None:
            lthr_filter = np.min(corr_matrix)
        if uthr_filter is None:
            uthr_filter = np.max(corr_matrix)

        coordinates = self.select_frame_coordinates(frame=frame)

        # prints
        print("@>: coordinates")
        print("@>: min value in correlation matrix: {0:.2f} Angstrom.".format(np.min(corr_matrix)))
        print("@>: max value in correlation matrix: {0:.2f} Angstrom.".format(np.max(corr_matrix)))
        print("@>: lower threshold: {0:.2f} Angstrom.".format(lthr_filter))
        print("@>: upper threshold: {0:.2f} Angstrom.".format(uthr_filter))

        print('@>: writing pymol file ... ')

        cmap_name = kwargs.get('cmap_name', 'coolwarm')
        cmap = plt.get_cmap(cmap_name)

        vmin = kwargs.get('vmin', 0)
        vmax = kwargs.get('vmax', 1)
        mpl.cm.ScalarMappable(norm=mpl.colors.Normalize(vmin=vmin, vmax=vmax, clip=False),
                              cmap=cmap).set_clim(vmin=vmin, vmax=vmax)

        cylinder_string = ("CYLINDER,  {0:.3f}, {1:.3f}, {2:.3f},{3:.3f}, {4:.3f}, {5:.3f}, {6:.3f}, " +
                           "{7:.3f}, {8:.3f}, {9:.3f}, {10:.3f}, {11:.3f}, {12:.3f}, \n ")

        spheres_string = "show spheres, chain {} and resi {} and {}\n"

        PML = open(pml_out_file + '_all.pml', 'w')
        PML.write(f"cgo_transparency, .4\n")
        PML.write(f"load {pdb_file} \n")
        PML.write("color gray40, polymer.protein\n")
        PML.write("set sphere_scale, 0.75\n")

        spheresList = []
        for i in range(0, len(corr_matrix)):
            for j in range(i + 1, len(corr_matrix)):
                if ((corr_matrix[i, j] > lthr_filter) and
                        (corr_matrix[i, j] <= uthr_filter)):
                    spheresList.append(i)
                    spheresList.append(j)

        print('@> atom_group_selstr', self.atom_group_selstr)

        for i, item in enumerate(np.unique(spheresList)):
            PML.write(spheres_string.format(ags.atoms.chainIDs[item], ags.atoms.resids[item], selection))

        PML.write("python\n")
        PML.write("from pymol.cgo import *\n")
        PML.write("from pymol import cmd\n")
        PML.write("correlations = [ \n")

        # Iterate over the elements of the matrix and plot a colored square for each element

        for i in range(0, len(corr_matrix)):
            for j in range(i + 1, len(corr_matrix)):
                if (corr_matrix[i, j] > lthr_filter) and (corr_matrix[i, j] <= uthr_filter):
                    color = cmap(corr_matrix[i][j])
                    color = list(color)
                    PML.write(cylinder_string.format(coordinates[i, 0], coordinates[i, 1], coordinates[i, 2],
                                                     coordinates[j, 0], coordinates[j, 1], coordinates[j, 2],
                                                     np.absolute(corr_matrix[i, j]) * edge_scaling,
                                                     color[0], color[1], color[2], color[0], color[1], color[2]))
        PML.write("]\n")
        PML.write("cmd.load_cgo(correlations,'correlations')\n")
        PML.write("cmd.set(\"cgo_line_width\",2.0,'correlations')\n")
        PML.write("python end")
        PML.close()

        chains = np.unique(ags.atoms.chainIDs)

        if (len(chains) > 1) & chainblocks:
            # write inter-chain correlations
            for chainI in chains:
                for chainJ in chains:
                    if chainI != chainJ:
                        PML = open(f"{pml_out_file}-interchain-chains-{chainI}-{chainJ}.pml", 'w')
                        PML.write(f"load {pdb_file} \n")
                        PML.write("cartoon type = tube\n")
                        PML.write("set sphere_scale, 0.75\n")
                        spheresList = []
                        for i in range(0, len(corr_matrix)):
                            for j in range(i + 1, len(corr_matrix)):
                                if (corr_matrix[i, j] > lthr_filter) and (corr_matrix[i, j] <= uthr_filter):
                                    if (ags.atoms.chainIDs[i] == chainI) and (ags.atoms.chainIDs[j] == chainJ):
                                        spheresList.append(i)
                                        spheresList.append(j)

                        for item in np.unique(spheresList):
                            PML.write(spheres_string.format(ags.atoms.chainIDs[item], ags.atoms.resids[item],
                                                            self.atom_group_selstr))
                        PML.write("python\n")
                        PML.write("from pymol.cgo import *\n")
                        PML.write("from pymol import cmd\n")
                        PML.write("correlations = [ \n")
                        for i in range(0, len(corr_matrix)):
                            for j in range(i + 1, len(corr_matrix)):
                                if (corr_matrix[i, j] > lthr_filter) and (corr_matrix[i, j] <= uthr_filter):
                                    if (ags.atoms.chainIDs[i] == chainI) and (ags.atoms.chainIDs[j] == chainJ):
                                        color = cmap(corr_matrix[i][j])
                                        color = list(color)
                                        PML.write(cylinder_string.format(coordinates[i, 0], coordinates[i, 1],
                                                                         coordinates[i, 2],
                                                                         coordinates[j, 0], coordinates[j, 1],
                                                                         coordinates[j, 2],
                                                                         np.absolute(corr_matrix[i, j]) * edge_scaling,
                                                                         color[0], color[1], color[2], color[0],
                                                                         color[1], color[2]))

                        PML.write("]\n")
                        PML.write("cmd.load_cgo(correlations,'correlations')\n")
                        PML.write("cmd.set(\"cgo_line_width\",2.0,'correlations')\n")
                        PML.write("python end")
                        PML.close()

            # write intra-chain correlations
            for chain in chains:
                PML = open(f"{pml_out_file}-intrachain-chain-{chain}.pml", 'w')
                PML.write(f"load {pdb_file} \n")
                PML.write("cartoon type = tube\n")
                PML.write("set sphere_scale, 0.75\n")
                spheresList = []
                for i in range(0, len(corr_matrix)):
                    for j in range(i + 1, len(corr_matrix)):
                        if (corr_matrix[i, j] > lthr_filter) and (corr_matrix[i, j] <= uthr_filter):
                            if (ags.atoms.chainIDs[i] == chain) and (ags.atoms.chainIDs[j] == chain):
                                spheresList.append(i)
                                spheresList.append(j)

                for item in np.unique(spheresList):
                    PML.write(spheres_string.format(ags.atoms.chainIDs[item], ags.atoms.resids[item],
                                                    self.atom_group_selstr))
                PML.write("python\n")
                PML.write("from pymol.cgo import *\n")
                PML.write("from pymol import cmd\n")
                PML.write("correlations = [ \n")
                for i in range(0, len(corr_matrix)):
                    for j in range(i + 1, len(corr_matrix)):
                        if (corr_matrix[i, j] > lthr_filter) and (corr_matrix[i, j] <= uthr_filter):
                            if (ags.atoms.chainIDs[i] == chain) and (ags.atoms.chainIDsa[j] == chain):
                                color = cmap(corr_matrix[i][j])
                                color = list(color)
                                PML.write(
                                    cylinder_string.format(coordinates[i, 0], coordinates[i, 1], coordinates[i, 2],
                                                           coordinates[j, 0], coordinates[j, 1], coordinates[j, 2],
                                                           np.absolute(corr_matrix[i, j]) * edge_scaling,
                                                           color[0], color[1], color[2], color[0], color[1], color[2]))

                PML.write("]\n")
                PML.write("cmd.load_cgo(correlations,'correlations')\n")
                PML.write("cmd.set(\"cgo_line_width\",2.0,'correlations')\n")
                PML.write("python end")
                PML.close()


def display_community(path, sys, view, community_lookup, color_dict, outpath):
    """
    Pymol function to display communities on secondary structure

    Parameters
    ----------
    path: str,
        path to pdb
    sys: str,
        name of pdb to load (without extension)
    view: set,
        orientation matrix copied from pymol `get_view()`
            example:   view = (\
            -0.611808956,   -0.140187785,    0.778481722,\
             0.387945682,    0.804496109,    0.449760556,\
            -0.689334452,    0.577175796,   -0.437813699,\
             0.000000000,    0.000000000, -219.895217896,\
            45.563232422,   57.541908264,   47.740921021,\
            85.014022827,  354.776367188,   20.000000000 )
    community_lookup: mdigest.CMTY.nodes_communities_collect object,
        collected output from MD trajectory community analysis
    color_dict: dict,
        dictionary with color names as keys and rgb codes as values
    outpath: str,
        where to save png, format should be outpath = '/path/to/png/', png will be saved as outpath_communities.png
    """
    from pymol import cmd
    cmd.viewport(1892, 1754)
    cmd.delete('all')
    cmd.load(path + sys + ".pdb", 'pdb')
    cmd.set_view(view)
    cmd.show('cartoon', 'pdb')

    for c, nodes in community_lookup['comm_nodes'].items():
        com = np.asarray(nodes) + 1
        com = np.asarray(com, dtype=str)
        com = list(com)
        selection = ('+'.join(com).strip(''))
        selection = 'resi ' + selection
        cmd.select('curr_res', 'pdb and ' + selection)
        cmd.set_color('color_%d' % c, color_dict[c])
        cmd.color('color_%d' % c, 'curr_res')

    cmd.set_view(view)
    cmd.enable("all")
    print('@>: RAYTRACE')
    # cmd.ray(800, 800)
    cmd.png(outpath + '_communities.png', ray=1)


cmd.extend('display_community', display_community)


def ss_network(ss_stats, gcc, nodes_save_path, edges_save_path, num_sd=1.5):
    """
    Secondary structures network

    Parameters
    ----------
    ss_stats: pd.DataFrame,
        dataframe of secondary structure information, obtained from self.ss_stats
    gcc: np.ndarray of shape (nfeatures*nfeatures),
        pairwise generalized correlation coefficients
    nodes_save_path: str,
        path to save dictionary of nodes
    edges_save_path: str,
        path to save dictionary of edges
    num_sd: int,
        minimum standard deviations above the mean value for an edge between nodes to be considered as significant

    Returns
    -------
    dictionary of nodes: dict,
    dictionary of edges: dict
    """

    # assign secondary structure elements based on probability
    ss_stats['SSE'] = 0
    for i in range(0, len(ss_stats)):
        if ss_stats['% Coil'][i] > 50:
            ss_stats['SSE'][i] = 'coil'
        if ss_stats['% Helix'][i] > 50:
            ss_stats['SSE'][i] = 'helix'
        if ss_stats['% Strand'][i] > 50:
            ss_stats['SSE'][i] = 'strand'
        if ss_stats['% Strand'][i] < 50 and ss_stats['% Helix'][i] < 50 and ss_stats['% Coil'][i] < 50:
            ss_stats['SSE'][i] = 'diverse'
    ss_list = list(ss_stats['SSE'][1:])

    # create dictionary containing secondary structure group assignments
    community = {}
    community[int(f'{0}')] = list()
    num = 0
    for i in range(0, len(ss_list)):
        if ss_list[i] != ss_list[i - 1]:
            num += 1
            community[int(f'{num}')] = list()
            community[int(f'{num}')].append(i)
        else:
            community[int(f'{num}')].append(i)

    # compute edges
    edges = {}
    for i in range(0, len(ss_list)):
        for j in range(0, len(ss_list)):
            edges[f'{i}, {j}'] = gcc[i][j]

    # delete self-correlation keys and values
    for k, v in edges.copy().items():
        if v == 2.0:
            del edges[f'{k}']
    data_array = np.array(list(edges.items()))

    # create dataframe containing nodes and edges
    dataframe = pd.DataFrame(data_array, columns=['nodes', 'correlation'])
    nodes = dataframe['nodes'].str.split(',', expand=True)
    nodes = nodes.rename(columns={0: 'node 1', 1: 'node 2'})
    correlation = pd.DataFrame(dataframe['correlation'])
    df = pd.merge(nodes, correlation, left_index=True, right_index=True)
    df = df.iloc[::2]
    df = df.reset_index(drop=True)
    df['node 1'] = df['node 1'].astype('int')
    df['node 2'] = df['node 2'].astype('int')
    df['correlation'] = df['correlation'].astype('float')
    df = df.reset_index(drop=True)

    # delete correlations < 2 sd's above mean
    cutoff = np.mean(df['correlation']) + 2 * np.std(df['correlation'])
    df = df[abs(df['correlation']) >= cutoff]
    df = df.reset_index(drop=True)

    # compute node 1 groups
    node1 = []
    for i in range(0, len(df)):
        for k, v in iter(community.items()):
            if df['node 1'][i] in community[k]:
                node1.append(k)
    df = pd.merge(df, pd.DataFrame(node1), left_index=True, right_index=True)
    df = df.rename(columns={0: 'node 1 community'})

    # compute node 2 groups
    node2 = []
    for i in range(0, len(df)):
        for k, v in iter(community.items()):
            if df['node 2'][i] in community[k]:
                node2.append(k)
    df = pd.merge(df, pd.DataFrame(node2), left_index=True, right_index=True)
    df = df.rename(columns={0: 'node 2 community'})

    # delete intra-group correlations
    df = df[df['node 1 community'] != df['node 2 community']]
    df = df.reset_index(drop=True)

    # sum correlations between each secondary structure group pair
    communication = {}
    for i in range(0, len(df)):
        node1 = df['node 1 community'][i]
        node2 = df['node 2 community'][i]
        if f'{node1}, {node2}' in communication:
            communication[f'{node1}, {node2}'] += df['correlation'][i]
        else:
            communication[f'{node1}, {node2}'] = df['correlation'][i]

    # delete correlations < num_sd's above mean
    cutoff = np.mean(list(communication.values())) + num_sd * np.std(list(communication.values()))
    data = {k: v for k, v in communication.items() if abs(v) >= cutoff}

    # define nodes of significant edges
    edge_list = [k for k, v in data.items()]
    for i in range(len(edge_list)):
        edge_list[i] = edge_list[i].split(',')

    list2 = []
    for i in range(len(edge_list)):
        for j in edge_list[i]:
            list2.append(int(j))

    nodes = []
    for i in range(len(list2)):
        if str(list2[i]) not in nodes:
            nodes.append(int(list2[i]))

    comm = {}
    for i in range(len(nodes)):
        for k, v in community.items():
            if nodes[i] == k:
                comm[nodes[i]] = v

    # write out nodes
    with open(nodes_save_path, 'w') as comm_dict:
        comm_dict.write(str(comm))

    # write out edges
    with open(edges_save_path, 'w') as edge_dict:
        edge_dict.write(str(data))


def draw_electrostatic_network(communities_path, edges_path, save_path, fetch_pdb=None, pse_path=None,
                               edge_multiplier=5, color_ss=True):
    """
    Draw electrostatic network

    Parameters
    ----------
    communities_path: str,
        path to communities txt file
    edges_path: str,
        path to edges text file
    save_path: str,
        path to save pse file
    fetch_pdb: str,
        PDB ID to fetch, default=None
    pse_path: str,
        path to structural file if fetch_pdb=None, default=None
    edge_multiplier: int,
        multiplicative factor for edge widths in visualization, default=5
    color_ss: bool,
        whether to color structure by secondary structure, default=True

    Returns
    -------
    .pse, PyMOL pse file,

    """
    from pymol import cmd
    # initialize pymol
    cmd.delete('all')
    # load structure
    if fetch_pdb != None:
        cmd.fetch(fetch_pdb, type='pdb1')
    else:
        cmd.load(pse_path)
    # open communities dictionary in string form and convert to dictionary
    with open(communities_path) as f:
        nodes = eval(f.read())
        print('NODES: ', nodes)
    # open edges dictionary in string form and convert to dictionary
    with open(edges_path) as f:
        edges = eval(f.read())
        print('EDGES: ', edges)
    # get nodes of significant edges
    edges_split = {}
    for k, v in edges.items():
        edges_split[k.split(', ')[0]] = v
        edges_split[k.split(', ')[1]] = v
    # define list of nodes in significant edges
    node_list = []
    for ke, ve in edges_split.items():
        node_list.append(ke)
    # define a selection for each community, rename selections
    groups = []
    for kn, vn in nodes.items():
        if str(kn) in node_list:
            cmd.select('resid ' + str(vn[0]) + '-' + str(vn[-1]))
            name0 = 'group' + str(kn)
            cmd.set_name('sele', name0)
            groups.append(name0)
    # set visualization parameters
    # show groups as surfaces
    cmd.show('surface', str(groups)[1:-1])
    # set surface transparency
    cmd.set('transparency', 0.6)
    # set cartoon transparency
    cmd.set('cartoon_transparency', 0.2)

    # define script to compute center of mass
    def com(selection, state=None, mass=None, object=None, quiet=1, **kwargs):
        quiet = int(quiet)
        if (object == None):
            try:
                object = cmd.get_legal_name(selection)
                object = cmd.get_unused_name(object + "_COM", 0)
            except AttributeError:
                object = 'COM'
        cmd.delete(object)
        if (state != None):
            x, y, z = get_com(selection, mass=mass, quiet=quiet)
            if not quiet:
                print("%f %f %f" % (x, y, z))
            cmd.pseudoatom(object, pos=[x, y, z], **kwargs)
            cmd.show("spheres", object)
        else:
            for i in range(cmd.count_states()):
                x, y, z = get_com(selection, mass=mass, state=i + 1, quiet=quiet)
                if not quiet:
                    print("State %d:%f %f %f" % (i + 1, x, y, z))
                cmd.pseudoatom(object, pos=[x, y, z], state=i + 1, **kwargs)
                cmd.show("spheres", 'last ' + object)

    cmd.extend("com", com)

    def get_com(selection, state=1, mass=None, quiet=1):
        """
        Calculate the center of mass

        ``[*]`` function adapted from: https://github.com/Pymol-Scripts/Pymol-script-repo/blob/master/center_of_mass.py
        """
        quiet = int(quiet)
        totmass = 0.0
        state = int(state)
        model = cmd.get_model(selection, state)
        x, y, z = 0, 0, 0
        for a in model.atom:
            if (mass != None):
                m = a.get_mass()
                x += a.coord[0] * m
                y += a.coord[1] * m
                z += a.coord[2] * m
                totmass += m
            else:
                x += a.coord[0]
                y += a.coord[1]
                z += a.coord[2]
        if (mass != None):
            return x / totmass, y / totmass, z / totmass
        else:
            return x / len(model.atom), y / len(model.atom), z / len(model.atom)

    cmd.extend("get_com", get_com)
    # compute center of mass for each group
    for i in range(len(groups)):
        com(str(groups[i]))
    # define positive edge distances
    for k, v in edges.items():
        if v > 0:
            cmd.distance('d' + k.split(', ')[0] + k.split(', ')[1], 'group' + k.split(', ')[0] + '_COM',
                         'group' + k.split(', ')[1] + '_COM')
            cmd.set('dash_width', v * edge_multiplier, 'd' + k.split(', ')[0] + k.split(', ')[1])
            cmd.color('green', 'd' + k.split(', ')[0] + k.split(', ')[1])
        if v < 0:
            cmd.distance('d' + k.split(', ')[0] + k.split(', ')[1], 'group' + k.split(', ')[0] + '_COM',
                         'group' + k.split(', ')[1] + '_COM')
            cmd.set('dash_width', v * edge_multiplier, 'd' + k.split(', ')[0] + k.split(', ')[1])
            cmd.color('red', 'd' + k.split(', ')[0] + k.split(', ')[1])
    # remove dash gap
    cmd.set('dash_gap', 0)
    # hide labels
    cmd.hide('labels')
    # hide com spheres
    cmd.hide('spheres')
    # remove solvent molecules
    cmd.remove('sol')
    if color_ss == True:
        cmd.color('lightblue', 'ss h')
        cmd.color('lightmagenta', 'ss s')
        cmd.color('gray50', 'ss l+')
    # save pse file
    cmd.save(save_path)



from mdigest.utils.imports import *

# Functions to print to screen
def print_screen_logfile(string, opened_logfile):
    """
    Print string on screen and write it on logfile

    Parameters
    ----------
    string: str,
        string to print and write
    opened_logfile: file
        opened logfile
    """

    opened_logfile.write("{}\n".format(string))
    opened_logfile.flush()  # forcing the writing by flushing the buffer


###############################################################################
###### Trajectory handling functions ##########################################
###############################################################################

# Functions to reduce trajectories
def reduce_trajectory(mda_u, selection, initial=0, final=-1, step=1):
    """
    Reduce the trajectory according to the `selection` string

    Parameters
    ----------
    mda_u: mda.Universe,
    selection: str
    initial: int,
    final: int or -1,
    step: int,

    Returns
    --------
    :returns reduced: mda.Universe object
        a reduced trajectory (mdAnalysis.Universe.trajectory)
    """

    atomgroup = mda_u.select_atoms(selection)

    ids = mda_u.select_atoms(selection).atoms.ids
    # initalize coordinates array
    if final == -1:
        final = len(mda_u.trajectory)
    coordinates = np.zeros((int(np.ceil((final - initial) / step)), len(atomgroup), 3))
    count = 0
    for ts in mda_u.trajectory[initial:final:step]:
        coordinates[count, :, :] = ts.positions[ids, :]
        count += 1
    # construct new universe with selected atoms only
    reduced = mda.Merge(atomgroup).load_new(coordinates[None, :, :], order="fac")
    return reduced


def reduce_mdt_trajectory(mdt_u, selection):
    """
    Reduce the trajectory according to the selection string

    Parameters
    ----------
    mdt_u: mdtraj.Trajectory,
    selection: str,

    Returns
    -------
    reduced, mdtraj.Trajectory object
        a reduced trajectory (mdtraj.Trajectory)
    """
    reduced = mdt_u.restrict_atoms(mdt_u.topology.select(selection))

    print('@>: reduced trajectory to %d atoms' % len(mdt_u.topology.atoms))

    return reduced

# Functions to write trajectories
def write_universe(mda_u, path, filename, chainid=False, initial=0, final=-1, step=1):
    """
    Write a trajectory and corresponding pdb to file

    Parameters
    ----------
    :params trajectory: mdAnalysis.Universe.trajectory,
    :params path: str,
    :params filename: str,
    :params chainid: bool,
        if True, write chainid to pdb
    :params ititial: int,
    :params final: int or -1
    :params step: int

    Returns
    -------
    reduced: mda.Universe object,
        a reduced trajectory (mdAnalysis.Universe.trajectory)
    """

    if final == -1:
        final = len(mda_u.trajectory)

    mda_u.trajectory[0]
    with mda.Writer(path + filename + '.dcd', mda_u.atoms.n_atoms) as W:
        for ts in mda_u.trajectory[initial:final:step]:
            W.write(mda_u.atoms)
    mda_u.trajectory[0]
    with mda.Writer(path + filename + '.pdb', mda_u.atoms.n_atoms) as W:
        if chainid:
            mda_u.add_TopologyAttr('chainIDs')
            mda_u.atoms.chainIDs = mda_u.atoms.segids
        W.write(mda_u.atoms)
    W.close()


"""#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: fmaschietto, bcallen95"""

from MDAnalysis.analysis.dihedrals import Ramachandran

import mdigest.core.auxiliary as aux
import mdigest.core.savedata as sd
import mdigest.core.toolkit as tk
from   mdigest.core.imports import *


class DihDynCorr:
    """Correlated motions of dihedrals"""
    # TODO CHANGE Nresidues to Natoms - to enable the calculation of correlation over different selections and more than one node per residue

    def __init__(self, MDSIM):
        """
         Description
         -----------
         General purpose class handling computation of different correlation metrics from dihedrals fluctuations sampled
         over MD trajectories. Each (selected) residue is described using the transformation --> [$sin(\phi)$, $cos(\phi)$, $sin(\psi)$, $cos(\psi)$]

         Parameters
         ----------
         MDSIM: class object
             DynCorr inherits general attributes from MDSIM

         Methods
         -------

         Attributes
         ----------
         self.nodes_dih_to_indices_dictionary: dict,
             nodes to indices dictionary; useful for later plotting (match dihedral nodes id to resid)
         self.dihedrals_allreplicas: dict, with replica index ``rep_n`` as key and values of shape (nsamples, nfeatures * features_dimension)
             being the projected dihedrals values array.
         self.disp_from_mean_dih_allreplicas: dict, with replica index ``rep_n`` as key and values of shape (nsamples, nfeatures)
             being the displacement of each atom from the average position computed over all the selected timesteps
         self.covar_dih_allreplicas: dict, with replica index ``rep_n`` as key and values of shape (nfeatures, nfeatures)
             being the covariance matrices of atomic displacements for each given trajectory replica
         self.dih_gcc_allreplicas: nested dict, with replica index ``rep_n`` as key and inner dict with key ``gcc_mi`` or ``gcc_lmi`` and values of shape (nfeatures, nfeatures)
             being the mutual information based generalized correlation coefficient matrix[1] for each replica. ``gcc_lmi`` linearized mutual information based generalized correlation using gaussian estimator; ``gcc_mi`` mutual information based generalized correlation computed using nonlinear estimator.
         self.dih_dcc_allreplicas: dict, with replica index ``rep_n`` as key and values of shape (nfeatures, nfeatures)
             being the normalized dynamical cross-correlations matrices for each replica
         self.dih_pcc_allreplicas: dict, with replica index ``rep_n`` as key and values of shape (nfeatures, nfeatures)
             being the Pearson's product-moment correlation coefficients for each replica.
         self.ramachandran: object,
             mda.Ramachandran output
         self.dih_values: np.ndarray,
             dihedrals array
         self.dih_labels:  list,
             dihedrals labels
         self.dih_indices: np.ndarray,
             array of indices corresponding to the selected dihedrals

         References
         ----------

         Examples
         --------
         """
        self.mds_data                = MDSIM.mds_data
        self.mda_u                   = MDSIM.mda_u
        self.atom_group_selstr       = MDSIM.atom_group_selstr
        self.system_selstr           = MDSIM.system_selstr
        self.atom_group_selection    = MDSIM.atom_group_selection
        self.nodes_idx_array         = MDSIM.nodes_idx_array
        self.nodes_to_res_dictionary = MDSIM.nodes_to_res_dictionary

        # ---------------------------------------------#
        self.natoms              = MDSIM.natoms
        self.total_nframes       = MDSIM.total_nframes
        self.nframes_per_replica = MDSIM.nframes_per_replica
        self.num_replicas        = MDSIM.num_replicas
        self.nresidues           = MDSIM.nresidues
        self.initial             = MDSIM.initial
        self.final               = MDSIM.final
        self.step                = MDSIM.step
        self.window_span         = MDSIM.window_span

        # ---------------------------------------------#
        self.ramachandran      = None
        self.dih_values        = None
        self.dih_labels        = None
        self.dih_indices       = None

        self.nodes_dih_to_indices_dictionary  = {}  # useful for later plotting (match dihedral nodes id to resid)
        self.dihedrals_allreplicas            = {}
        self.disp_from_mean_dih_allreplicas   = {}
        self.covar_dih_allreplicas            = {}
        self.dih_gcc_allreplicas              = {}
        self.dih_dcc_allreplicas              = {}
        self.dih_pcc_allreplicas              = {}
        self.dih_eigen_centrality_allreplicas = {}

    def save_class(self, file_name_root='./output/cache/'):
        """
        Save DihDynCorr class instances to file

        file_name_root: srt
            path where to save class
        """

        self.mds_data = sd.MDSdata()
        self.mds_data.nodes_dih_to_indices_dictionary = self.nodes_dih_to_indices_dictionary
        self.mds_data.nodes_to_res_dictionary = self.nodes_to_res_dictionary
        self.mds_data.atom_group_selstr = self.atom_group_selstr
        self.mds_data.system_selstr = self.system_selstr
        self.mds_data.dih_labels = self.dih_labels
        self.mds_data.dih_indices = self.dih_indices
        self.mds_data.dihedrals_allreplicas = self.dihedrals_allreplicas
        self.mds_data.disp_from_mean_dih_allreplicas = self.disp_from_mean_dih_allreplicas
        self.mds_data.covar_dih_allreplicas = self.covar_dih_allreplicas
        self.mds_data.dih_gcc_allreplicas = self.dih_gcc_allreplicas
        self.mds_data.dih_dcc_allreplicas = self.dih_dcc_allreplicas
        self.mds_data.dih_pcc_allreplicas = self.dih_pcc_allreplicas
        self.mds_data.dih_eigen_centrality_allreplicas = self.dih_eigen_centrality_allreplicas
        self.mds_data.save_to_file(file_name_root)


    def parse_dih_dynamics(self, mean_center=False, LMI='gaussian', MI='knn_5_1', DCC=False, PCC=False, COV_DISP=True, CENTRALITY=True, **kwargs):
        """
        General purpose class handling computation of different correlation metrics from $\phi$, \$psi$ backbone dihedrals fluctuations sampled over MD trajectories.
        Diedrals are transformed using $\phi$ --> {$sin(\phi)$, $cos(\phi)$} and $\psi$ --> {$sin(\psi)$, $cos(\psi)$} such that each residue
        (temimal residues excluded) is described by an array of four entries [$sin(\phi)$, $cos(\phi)$, $sin(\psi)$, $cos(\psi)$].

        Parameters
        ----------
        mean_center: bool
            wheter to subtract mean
        LMI: str or None; default 'gaussian'
            - 'gaussian' for using gaussian estimator
            - None: skip computation of linearized mutual information based correlation

        MI: str or None, default 'knn_5_1'
            composite argument where knn specifiess use of k-nearest neighbor algorithm,
            5 specifies number of nearest neighbours, 1 specifies estimate to use (options are 1 or 2)
        DCC: bool,
            whether to compute dynamical cross correltaion
        PCC: bool,
            whether to compute Pearson's cross correlation
        COV_DISP: bool,
            whether to compute covariance of dihedrals displacements
        CENTRALITY: bool,
            whether to compute eigen centrality of dihedrals fluctuations. Default is True
        kwargs:
            - normalized: bool
                whether to normalize DCC matrix
            - subset: list,
                list of indices specifying the nodes for which to compute MI
            - center: str or None
                How to compute the covariance matrix; possible values are 'mean' or 'square_disp'
            - reference_frame: bool,
                Standardize dihedrals by centering distribution to a frame of reference
            - median_center: bool,
                Standardize dihedrals by centering distribution to the median

        Returns
        -------

        """

        # setup kwargs
        if DCC:
            try:
                normalized = kwargs['normalized']
            except KeyError:
                print('@>: DCC normalization is set to True. Use normalize=False to avoid normalization.')
                normalized = True

        if COV_DISP:
            try:
                center = kwargs['center']  # None, mean or square_disp
            except KeyError:
                print('@>: Use center to select different ways to compute the covariance'
                      'possible options are None, mean or square_disp')
                center = 'square_disp'
                print('@>: center  = ', center)

        # setup kwargs
        try:
            subset = kwargs['subset']
        except KeyError:
            subset = None

        try:
            reference_frame = kwargs['reference_frame']
        except KeyError:
            reference_frame = 0

        try:
            median_center = kwargs['median_center']
        except KeyError:
            median_center = False

        try:
            reference_to_frame = kwargs['']
        except:
            reference_to_frame = False

        solvers = []
        if MI is not None:
            solvers.append(MI)
        if LMI is not None:
            solvers.append(LMI)

        num_replicas = self.num_replicas

        for win_idx in tk.log_progress(range(self.num_replicas), 1, size=num_replicas, name="Window"):

            beg = self.initial + self.window_span * win_idx
            end = self.initial + self.window_span * (win_idx + 1)

            stride = self.step
            print("@>: start, end frames:", beg, end)

            print("@>: Dihedrals calculation ...")
            self.ramachandran = Ramachandran(self.mda_u.universe.select_atoms(self.system_selstr)).run()
            r = self.ramachandran

            if reference_to_frame:
                # reference to selected frame
                dih_reference = r.angles[reference_frame, :]
                cos = np.cos(np.radians(r.angles - dih_reference))
                sin = np.sin(np.radians(r.angles - dih_reference))
                self.dih_values = np.concatenate([cos, sin], axis=2)
                mean_center = False

            if median_center:
                # subtract median
                median = np.median(r.angles, axis=0)
                cos = np.cos(np.radians(r.angles - median))
                sin = np.sin(np.radians(r.angles - median))
                self.dih_values = np.concatenate([cos, sin], axis=2)
                mean_center = False
                reference_to_frame = False

            if mean_center:
                # if no referencing to initial or other frame, and mean_center is True center with respect to average
                # mean_angles = np.mean(r.angles.mean(axis=0)[beg:end:stride, :], axis=0)
                mean_angles = np.mean(r.results.angles, axis=0)

                cos = np.cos(np.radians(r.results.angles - mean_angles))
                sin = np.sin(np.radians(r.results.angles - mean_angles))
                self.dih_values = np.concatenate([cos, sin], axis=2)

            else:
                #plain dihedrals
                cos = np.cos(np.radians(r.results.angles))
                sin = np.sin(np.radians(r.results.angles))
                self.dih_values = np.concatenate([cos, sin], axis=2)


            self.dih_labels = ['cos_phi', 'cos_psi', 'sin_phi', 'sin_psi']
            self.dih_indices = np.stack([r.ag1.residues.resindices, r.ag4.residues.resindices], axis=-1)
            features_dim = len(self.dih_labels)

            # use first column of dih_indices for mapping
            self.nodes_dih_to_indices_dictionary = dict(
                zip(self.dih_indices[:, 0], [res.phi_selection().atoms.ids for res
                                             in self.mda_u.residues if
                                             res.phi_selection() is not None]))

            print("@>: LMI calculation ...")
            print("@>: begin frame: %d" % beg)
            print("@>: end   frame: %d" % end)
            print("@>: step:        %d" % self.step)

            values = self.dih_values[beg:end:stride, :, :]
            print("@>: dih matrix shape: {}".format(values.shape))

            self.dihedrals_allreplicas['rep_%d' % win_idx] = values

            MIdict = {}
            ECdict = {}
            if solvers is not []:
                for solver in solvers:
                    if solver == 'gaussian':
                        print("@>: computing lmi correlation matrix")
                        MIdict.update({'gcc_lmi': aux.compute_generalized_correlation_coefficients(values,
                                                                                                   features_dimension=len(
                                                                                                       self.dih_labels),
                                                                                                   solver=solver,
                                                                                                   correction=False)})
                        if CENTRALITY:
                            print("@>: computing eigenvector centrality from lmi matrix")
                            _, ec = aux.compute_eigenvector_centrality(MIdict['gcc_lmi'], weight='weight')
                            ECdict.update({'gcc_lmi': ec})

                    elif 'knn' in solver:
                        print("@>: computing mi correlation matrix")
                        MIdict.update({'gcc_mi': aux.compute_generalized_correlation_coefficients(values,
                                                                                                  features_dimension=len(
                                                                                                      self.dih_labels),
                                                                                                  solver=solver,
                                                                                              correction=True)})
                        if CENTRALITY:
                            print("@>: computing eigenvector centrality from mi matrix")
                            _, ec = aux.compute_eigenvector_centrality(MIdict['gcc_mi'], weight='weight')
                            ECdict.update({'gcc_mi': ec})

                self.dih_gcc_allreplicas['rep_%d' % win_idx] = MIdict
                self.dih_eigen_centrality_allreplicas['rep_%d' % win_idx] = ECdict

            if COV_DISP:
                print("@>: computing covariance of dihedral fluctuations")
                self.covar_dih_allreplicas['rep_%d' % win_idx] = aux.evaluate_covariance_matrix(values, center=center)

            if PCC:
                print("@>: computing square displacements")
                self.disp_from_mean_dih_allreplicas['rep_%d' % win_idx] = aux._compute_square_displacements(values)

                print("@>: computing pearson correlation matrix")
                self.dih_pcc_allreplicas['rep_%d' % win_idx] = \
                    np.corrcoef(self.disp_from_mean_dih_allreplicas['rep_%d' % win_idx], rowvar=False)

            if DCC:
                print("@>: computing DCC")
                values_reshaped = values.reshape((values.shape[0], values.shape[1] * values.shape[2]))
                values_ave = np.mean(values_reshaped, axis=0)
                self.dih_dcc_allreplicas['rep_%d' % win_idx] = aux.compute_DCC(values_reshaped, features_dim,
                                                                               normalized=normalized)



"""#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: fmaschietto, bcallen95"""

from mdigest.core.imports import *
from sklearn.neighbors import KDTree
from scipy.special import digamma
import numba

##############################################################################
################# Correlation related functions #############################
##############################################################################
def _extract_coords(u, selection='protein'):
    """
    Utility function to extract coordinates from an MDA.Universe given some atom selecton.

    Parameters
    ----------
    u: mda.Universe object,
        MDAnalysis universe
    selection: str,
        atom selection string

    Returns
    -------
    all_coords: np.ndarray
        coordinates
    """
    nfs = u.trajectory.n_frames
    subset = u.select_atoms(selection)
    all_coords = np.zeros(shape=(nfs, len(subset), 3))
    counter = 0
    for ts in u.trajectory:
        temp_pos = subset.atoms.positions
        all_coords[counter] = temp_pos
        counter += 1
    return all_coords


def coordinate_reshape(values):
    """
    Reshape values to shape (nframes, nfeatures*features_dimensions) where
         - ``nframes`` is the number of frames of the MD timeseries considered,
         - ``nfeatures`` is the number of features, i.e. number of atoms for which values are stored,
         - ``features_dimensions`` is the dimensionality of the features (i.e. 3 for xyz-coordinates, 4 for [sin(phi), cos(phi), sin(psi), cos(psi)] dihedral coordinate vector).

    Parameters
    ----------
    values: np.ndarray, shape (nframes, nfeatures, features_dimensions)
        values


    Returns
    -------
    values_: np.ndarray, shape (nframes, nfeatures * features_dimensions)
        reshaped values

    """

    nframes = values.shape[0]
    nfeatures = values.shape[1]
    features_dimension = values.shape[2]

    values_ = values.reshape((nframes, nfeatures * features_dimension))

    for i in range(0, nframes):
        ccount = 0
        for j in range(0, nfeatures):
            values_[i, ccount + 0] = values[i, j, 0]
            values_[i, ccount + 1] = values[i, j, 1]
            values_[i, ccount + 2] = values[i, j, 2]
            ccount += features_dimension
    ccount = 0

    return values_


def _compute_displacements(values):
    """
    Compute atomic displacements

    Parameters
    ----------
    values: np.ndarray, shape``nframes, nfeatures, features_dimensions``
        contains features coordinate vectors, for example xyz-coordinates or dihedral vectors.
         - ``nframes`` is the number of frames of the MD timeseries considered,
         - ``nfeatures`` is the number of features, i.e. number of atoms for which values are stored,
         - ``features_dimensions`` is the dimensionality of the features (i.e. 3 for xyz-coordinates, 4 for [sin(phi), cos(phi), sin(psi), cos(psi)] dihedral coordinate vector).

    Returns
    -------
    displacements: np.ndarray, shape (nframes, nfeatures * features_dimensions)
        displacements from mean position
    """

    nframes             = values.shape[0]
    nfeatures           = values.shape[1]
    features_dimensions = values.shape[2]

    values_ = values.reshape((nframes, nfeatures * features_dimensions))
    average = values_.mean(axis=0)

    displacements = np.zeros((nframes, nfeatures * features_dimensions))

    for i in range(0, features_dimensions * nfeatures):
        for j in range(0, nframes):
            displacements[j][i] = values_[j][i] - average[i]
    return displacements


def _compute_square_displacements(values):
    """
    Compute square displacements from features array.

    Parameters
    ----------
    values: np.ndarray,
        values of shapenframes, nfeatures, features_dimensions

    Returns
    -------
    displacements: np.ndarray, shape(nframes, nfeatures*1)
        displacements matrix
    """

    nframes             = values.shape[0]
    nfeatures           = values.shape[1]

    displacements = np.zeros((nframes, nfeatures))
    average = values.mean(axis=0)

    for j in range(0, nframes):
        displacements[j] = np.linalg.norm(values[j] - average, axis=1)
    return displacements


def compute_distance_matrix(values, features_dimensions):
    """
    Compute distance matrix using the average position over the entire lentgh of the given trajectory

    Parameters
    ----------
    values: np.ndarray,
        array of values of shape (nframes, nfeatures * features_dimensions)
    features_dimensions: int,
        dimensionality of features (3 if values contains xyz coordinates)

    Returns
    -------
    dist_matrix: np.ndarray, shape (nfeatures, nfeatures)
        distance matrix
    """

    from scipy.spatial import distance_matrix
    nfeatures = int(values.shape[1]/features_dimensions)
    values_avg = np.mean(values.view(), axis=0)
    values_cop = np.reshape(values_avg, [nfeatures, features_dimensions])
    dist_matrix = distance_matrix(values_cop, values_cop)
    return dist_matrix


def evaluate_covariance_matrix(values, center='square_disp'):
    """
    Compute the covariance matrix of features values.

    Parameters
    ----------
    values: np.ndarray, shape (nframes, nfeatures, features_dimensions)
        matrix of displacements from mean values
    center: str,
        -if 'mean' remove mean
        -if 'square_disp' compute square displacements of values array (default option)

    Returns
    -------
    covar_mat: np.ndarray, shape (nfeatures, nfeatures)
        covariance matrix
    """

    nframes             = values.shape[0]
    nfeatures           = values.shape[1]
    features_dimensions = values.shape[2]

    if center == 'mean':
        values_ = _compute_displacements(values)

    elif center == 'square_disp':
        values_ = _compute_square_displacements(values)


    else:
        values_ = values.reshape((nframes, nfeatures * features_dimensions))

    covar_mat = np.cov(values_, rowvar=False, bias=True)

    return covar_mat



def _mutual_information_gaussian_estimator(values, features_dimension, correction=False):
        """
        Computes linearized mutual information (LMI), based on Gaussian estimator.
            Unlike Pearson correlation, LMI detects orthogonal correlation but still does not capture non linear correlation.
            Allows for rapid performance with results that in most cases compare very well to those obtained using
            non-linear estimators of MI.

        Parameters
        ----------
        values: np.array,
            values should have shape (n_samples, n_features, feature_dimension)
        features_dimension: int,
            features shape(i.e. 3 for xyz coordinates)

        Returns
        -------
        mutual_information: np.ndarray, shape (nfeatures, nfeatures)
            linearized mutual information matrix
        """

        n_features = values.shape[1]

        if features_dimension == 1:
            print('@>: values array has shape {}'.format(values.shape))
            rho = np.corrcoef(values, rowvar=False)
            np.fill_diagonal(rho, 0)
            mutual_information = -0.5 * np.emath.log((1 - rho ** 2))
            if correction:
                mutual_information -= np.min(mutual_information)

        else:
            print('@>: values array has shape {}'.format(values.shape))
            Sx = np.stack([np.cov(elt) for elt in values.transpose(1, 2, 0)], axis=0)
            det = np.log(np.linalg.det(Sx))
            HxHy = np.add.outer(det, det)
            Hxy = np.zeros_like(HxHy)
            np.fill_diagonal(HxHy, 0)
            iterator = list(cwr(range(n_features), 2))

            for ij in tqdm(iterator):
                i, j = ij
                covs = np.cov(np.concatenate([values[:, i, :], values[:, j, :]], axis=-1).T)
                _det = np.linalg.det(covs)
                Hxy[[i, j], [j, i]] = _det

            np.fill_diagonal(Hxy, 1)
            Hxy = np.log(Hxy)
            mutual_information = 1 / 2 * (HxHy - Hxy)
            if correction:
                mutual_information -= np.min(mutual_information)
        return mutual_information


def _digamma_n(x, e, offset=1):
    """
    Computes the average digamma of the number of points between x at a distance e. Used in the k-nn estimator for correlation.

    See Also
    --------
    ``[**]`` function adapted from https://github.com/agheeraert/pmdlearn/blob/d554d4f02f46c0f5df4428a2a04549397cf13244/src/pmdlearn/features.py
    """

    tree = KDTree(x, metric='chebyshev')
    # We need to cheat using e-1e-10 so that we don't count kth neighbor
    # and finding in which dimension(s) the kth neighbor reaches e is too
    # expensive
    num_points = tree.query_radius(x, e - 1e-10, count_only=True)

    # Removing a count for self only if it is not null
    num_points = num_points[num_points.nonzero()] - 1

    # There's an extra count in num_points so we take it into account
    num_points += offset
    return digamma(num_points)


def _mutual_information_knn_estimator(values, features_dimension=1, k=5, estimate=1,
                                      correction=True, subset=None):
    """
    Knn estimator to evaluate mutual information. Is intrinsically non-linear but has long computation time.
    Implemetation of equation (8) and (9) in Kraskov et al., PHYSICAL REVIEW E 69, 066138 (2004).
    ``[**]``

    Parameters
    ----------
    values: np.array,
        values should be in format (n_samples, n_features, feature_dimension)
    features_dimension: int
        features dimensions (i.e. 3 for xyz coordinates)
    k: int
        nearest neighbours
    estimate: int (1 or 2)
        - offset=1, constant=0
        - offset=0, constant=1/k
    correction
        whether to apply correction (subtraction of min value of mutual information)
    subset: np.array (or None)
        indices of features for which to calculate MI, if None, all features are kept

    Returns
    -------
    mutual_information: np.ndarray, shape (nfeatures, nfeatures)
        mutual information matrix

    See Also
    --------
    ``[**]`` function adapted from https://github.com/agheeraert/pmdlearn/blob/d554d4f02f46c0f5df4428a2a04549397cf13244/src/pmdlearn/features.py
    """

    # The precise knn estimator changes one offset in the digamma function
    # and adds a constant in some cases.

    if estimate == 1:
        print('@>: estimate == 1')
        offset = 1
        costant = 0
    elif estimate == 2:
        print('@>: estimate == 2')
        offset = 0
        costant = -1 / k

    elif not (estimate == 1 or estimate == 2):
        print('@>: defaulting to estimate == 1')
        offset = 1
        costant = 0

    n_features = values.shape[1]

    # Initializing mutual information
    mutual_information = np.ones((n_features, n_features))
    mutual_information *= digamma(k) + digamma(values.shape[0]) + costant

    if subset is None:
        iterator = list(cwr(range(n_features), 2))
    else:
        iterator = subset

    if features_dimension == 1:
        values = np.asarray(values)[:, :, np.newaxis]

    for ij in tqdm(iterator):
        i, j = ij
        x = values[:, i, :]
        y = values[:, j, :]

        points = np.concatenate([x, y], axis=1)
        # Building KDTree with max norm metric
        tree = KDTree(points, metric='chebyshev')
        # Adding an offset to not count self as nearest neighbor
        e = tree.query(points, k=k + 1)[0][:, k]
        digamma_nxny = (_digamma_n(x, e, offset=offset) +
                        _digamma_n(y, e, offset=offset))
        digamma_nxny = digamma_nxny[np.isfinite(digamma_nxny)]
        mutual_information[[i, j], [j, i]] -= np.average(digamma_nxny)

    if correction:
        mutual_information -= np.min(mutual_information)
    return mutual_information


def compute_generalized_correlation_coefficients(values, features_dimension=1, solver='gaussian', correction=True, subset=None):
    """
    Rescale mutual information based correlation (`gcc`) between 0 and 1

    Parameters
    ----------
    values: np.ndarray, shape (nsamples, nfeatures, features_dimension)
         time-series values
    features_dimension: int,
        dimensionality of features array
    solver: string,
        which solver to use, default 'gaussian'
    correction: bool,
         whether to apply correction (subtraction of min value of mutual information)
    subset: np.array or None,
        indices of features for which to calculate `gcc`, if None, all features are kept

    Returns
    -------
    gcc: np.ndarray,
        generalized correlation coefficient matrix
    """

    MI = None
    if solver == 'gaussian':
        MI = _mutual_information_gaussian_estimator(values, features_dimension=features_dimension, correction=correction)

    elif 'knn' in solver:
        # Knn estimator is a string with two arguments encoded: knn_arg1_arg2,
        # arg1 is an integer representing the number of nn
        # arg2 is an integer that encodes the estimator used (1 or 2)
        k = int(solver[4])
        if len(solver) <= 6:
            estimate = 1
        else:
            estimate = int(solver[6])
        MI = _mutual_information_knn_estimator(values=values, features_dimension=features_dimension, k=k, estimate=estimate, correction=correction, subset=subset)

    gcc = np.sqrt(1 - np.exp(-2 * MI / features_dimension))
    np.fill_diagonal(gcc, 1)
    return gcc


@numba.njit
def compute_DCC_matrix(values, values_avg, features_dimension):
    """
    Compute dynamical cross-correlation matrix (upper triangle)
    ``[***]``

    Parameters
    ----------
    values: np.ndarray, shape (nsamples, nfeature, features_dimension)
        time-series values
    values_avg: np.ndarray, shape (nfeature, features_dimension)
        mean time-series values (values averaged over time)
    features_dimension: int,
        dimensionality of features array

    Returns
    -------
    cross_correlation: np.ndarray, shape (nfeatures, nfeatures)
        dynamical cross correlation matrix


    See Also
    --------
    ``[***]`` function adapted from https://github.com/tekpinar/correlationplus/blob/master/correlationplus/calculate.py
    """

    nsamples  = values.shape[0]
    nfeatures = int(values.shape[1]/features_dimension)

    cross_correlation = np.zeros((nfeatures, nfeatures), dtype=np.double)
    for k in range(0, len(values)):
        disp_ = np.subtract(values[k], values_avg)
        for i in range(0, nfeatures):
            idx_xyz_i = features_dimension * i
            for j in range(i, nfeatures):
                idx_xyz_j = features_dimension * j
                if features_dimension == 3:
                    cross_correlation[i][j] += (disp_[idx_xyz_i] * disp_[idx_xyz_j] +
                                                disp_[idx_xyz_i + 1] * disp_[idx_xyz_j + 1] +
                                                disp_[idx_xyz_i + 2] * disp_[idx_xyz_j + 2])
                elif features_dimension == 4:
                    cross_correlation[i][j] += (disp_[idx_xyz_i] * disp_[idx_xyz_j] +
                                                disp_[idx_xyz_i + 1] * disp_[idx_xyz_j + 1] +
                                                disp_[idx_xyz_i + 2] * disp_[idx_xyz_j + 2] +
                                                disp_[idx_xyz_i + 3] * disp_[idx_xyz_j + 3])
                elif features_dimension == 2:
                    cross_correlation[i][j] += (disp_[idx_xyz_i] * disp_[idx_xyz_j] +
                                                disp_[idx_xyz_i + 1] * disp_[idx_xyz_j + 1])
                elif features_dimension == 1:
                    cross_correlation[i][j] += (disp_[idx_xyz_i] * disp_[idx_xyz_j] )
                else:
                    print('cross_correlation cannot be computed for the specified features dimension')
                    break

    return cross_correlation

def compute_DCC(values_, features_dimension, normalized=True):
    """
    Compute normalized dynamical pairwise cross-correlation coefficients for each given replica

    Parameters
    ----------
    values_: np.ndarray, shape (nframes, natoms*features_dimensions)
        coordinates array
    features_dimension: int,
        dimension of features array
    normalized: bool,
        whether to normalize cross-correlation matrix; default is True

    Returns
    -------
    cross_correlation: np.ndarray, shape (natoms, natoms)
        cross-correlation matrix
    """

    # number of frames
    num_frames = values_.shape[0]
    nfeatures = int(values_.shape[1] / features_dimension)

    # exclude last entry of coordinates array - it contains the mean.
    values_average = np.mean(values_, axis=0)

    # compute DCC matrix
    cross_correlation = compute_DCC_matrix(np.asarray(values_, np.double),
                                            np.asarray(values_average, np.double), features_dimension)

    # Do the averaging
    cross_correlation = cross_correlation / float(num_frames)

    if normalized:
        # normalize
        norm_cross_correlation = np.zeros((nfeatures, nfeatures), dtype=np.double)
        for i in range(0, nfeatures):
            for j in range(i, nfeatures):
                norm_cross_correlation[i][j] = cross_correlation[i][j] / (
                            (cross_correlation[i][i] * cross_correlation[j][j]) ** 0.5)
                norm_cross_correlation[j][i] = norm_cross_correlation[i][j]
        cross_correlation = norm_cross_correlation
        return cross_correlation

    else:
        for i in range(0, nfeatures):
            for j in range(i + 1, nfeatures):
                cross_correlation[j][i] = cross_correlation[i][j]
        return cross_correlation



def corr(mat):
    """
    Compute pearson correlation

    Parameters
    ----------
    mat: np.ndarray,
        is an input matrix of shape (nsamples, nfeatures)

    Returns
    -------
    corrmat: np.ndarray, shape (nfeatures, nfeatures)
        pearson correlation matrix
    """
    corrmat = np.corrcoef(mat, rowvar=False)
    np.fill_diagonal(corrmat, 1)
    return corrmat


def prune_adjacency(mat, distmat, loc_factor=5.0, greater=False, lower=False):
    """
    Zero out adjacency matrics value according to the locality factor. By defalult entries corresponding to atom pairs at a distance equal or greater than the locality_factor (loc_factor) are set to zero.

    Parameters
    ----------
    mat: np.ndarray
        adjacency matrix
    distmat: np.ndarray
        distance matrix
    loc_factor: float x
        locality_factor, defines the distance threshold to use for pruning
    greater: bool
        whether to prune residues pairs at distances GREATER than the locality factor, default is False
    lower: bool
        whether to prune residue pairs at distances LOWER than the locality factor, default is False

    Returns
    --------
    mat: np.ndarray,
        pruned ajacency matrix
    """

    if greater:
        mat[distmat > loc_factor] = 0.
    elif lower:
        mat[distmat < loc_factor] = 0.
    else:
        print('@>: prune values greater than %d by default' % loc_factor)
        mat[distmat > loc_factor] = 0.

    return mat


def filter_adjacency(mat, distmat, loc_factor):
    """
    Filter the adjacency matrix using exponential to dump long-range contacts (and hence focus on short-range ones)

    Parameters
    ----------
    mat: np.ndarray,
        matrix
    distmat: np.ndarray,
        matrix based on which to prune mat
    loc_factor: float,
        locality factor (in Amstrongs), suggested 5 Å

    Returns
    -------
    adj_matx: np.ndarray,
        filtered matrix
    """

    distmat = distmat
    adj_matx = np.zeros(mat.shape)
    for i in range(0, len(mat)):
        adj_matx[i, i] = 0.0
        distmat[i, i] = 0.0
        for j in range(0, len(mat)):
            adj_matx[i, j] = mat[i, j] * np.exp(-1.0 * distmat[i, j] / loc_factor)
            adj_matx[j, i] = adj_matx[i, j]
    return adj_matx



def filter_adjacency_inv(mat, distmat, loc_factor):
    """
    Filter the adjacency matrix using exponential to dump short-range contacts
    (and hence focus on long-range ones)

    Parameters
    ----------
    mat: np.ndarray,
        matrix
    distmat: np.ndarray,
        matrix based on which to prune mat
    loc_factor: float,
        locality factor (in Amstrongs), suggested 5 Å

    Returns
    -------
    adj_matx: np.ndarray,
        filtered matrix
    """

    distmat = distmat
    adj_matx = np.zeros(mat.shape)
    for i in range(0, len(mat)):
        adj_matx[i, i] = 0.0
        distmat[i, i] = 0.0
        for j in range(0, len(mat)):
            adj_matx[i, j] = mat[i, j] * np.exp(-1.0 * loc_factor/ distmat[i, j])
            adj_matx[j, i] = adj_matx[i, j]
    return adj_matx


def reduce_trajectory(universe, segIDs):
    """
    It can be useful to work on a reduced trajectory without hydrogens, especially when visualizing stuff.
    This function can be used to generate a reduced MDAnalysis universe.

    Parameters
    ----------
    universe: MDAnalysis.core.Universe object
    segIDs: str,
        segIDs of groups of atoms to keep

    Returns
    -------
    reduced: MDAnalysis.core.Universe object
        reduced universe
    """

    from MDAnalysis.analysis.base import AnalysisFromFunction as mdaAFF
    from MDAnalysis.coordinates.memory import MemoryReader as mdaMemRead


    print("@>: the initial universe had {} atoms.".format(len(universe.atoms)))

    # initial selection
    sel_str = "segid " + " ".join(segIDs)
    group_sel = universe.select_atoms(sel_str)
    group_sel_noh = group_sel.select_atoms("not type H*")

    # merging a selection from the universe returns a new (and smaller) universe
    reduced = mda.core.universe.Merge(group_sel_noh)

    print("@>: the final universe has {} atoms.".format(len(reduced.atoms)))

    # load the new universe to memory, with coordinates from the selected residues
    print("@>: loading universe to memory...")

    reduced.load_new(mdaAFF(lambda ag: ag.positions.copy(),
                            group_sel_noh).run().results, format=mdaMemRead)
    return reduced


def linfit(x, y):
    """
    Return R^2 where x and y are array-like
    """

    slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(x, y)
    return slope, intercept, r_value**2


def create_graph(adj_matx):
    """
    Create nx.Graph instance from a np.ndarray adjacency_matrix
    """
    G = nx.Graph(adj_matx)
    G_ = G.copy().to_undirected()
    G_.remove_edges_from(nx.selfloop_edges(G_))
    G = G_.copy()
    return G


def vec_query(arr, my_dict):
    """
    Convert dictionary to array
    """
    return np.vectorize(my_dict.__getitem__)(arr)



def compute_eigenvector_centrality(mat, loc_factor=None, distmat=None, weight='weight'):
    """
    Compute eigenvector centrality based on a given correlation matrix and pairwise distance matrix

    Parameters
    ----------
    mat: np.ndarray,
        adiacency (correlation matrix) to diagonalize
    loc_factor: float,
        locality factor (filtering threshold)
    distmat: np.ndarray,
        distance matrix
    weight: str or None,
        if None treat adjaccency as binary, if 'weight' use values weights.

    Returns
    --------
    cdict: dict,
        centrality dictionary with nodes as keys and centrality coefficients as values
    cvec: np.ndarray,
        eigenvector centrality coefficients
    """

    if loc_factor is not None:
        adj_matx = filter_adjacency(mat, loc_factor=loc_factor, distmat=distmat)
    else:
        adj_matx = mat

    G = create_graph(adj_matx)
    # calculate all node centrality values in the system.
    cdict = nx.eigenvector_centrality(G, max_iter=10000, weight=weight)
    all_nodes = np.array(list(cdict.keys()))
    cvec = vec_query(all_nodes, cdict)

    return cdict, cvec


def sorted_eig(A):
    """
    Sort eigenvalues from larges to smallest and reorder eigenvectors accordingly

    Parameters
    -----------
    A: array, shape (nfeatures, nfeatures)
                array of eigenvalues

    Returns
    -------
     eigenValues: np.ndarray, shape (nfeatures)

     eigenVectors: np.ndarray, shape (nfeatures,nfeatures)
    """

    eigenValues, eigenVectors = scipy.linalg.eigh(A)
    idx = eigenValues.argsort()[::-1]
    eigenValues = eigenValues[idx]
    eigenVectors = eigenVectors[:, idx]

    return eigenValues, eigenVectors


def to_pickle(dataframe, output):
    """
    Dump dataframe to file using pickle

    Parameters
    -----------
    dataframe: pd.DataFrame,
        dataframe to pickle
    output: str,
        output name with path
    """

    pickle.dump(dataframe, open(output, 'wb'))


def get_centrality_df(input_class, cent_don, cent_acc, cent, selection):
    """
    Create dataframe containing eigenvector centrality from KS energy and from alpha carbon displacements

    Parameters
    ----------
    input_class: object,

    cent_don: np.ndarray,

    cent_acc: np.ndarray,

    cent: np.ndarray,

    selection: str,

    Returns
    -------
    temp_df:  pd.DataFrame()
        data frame containing electrostatic-centrality and CA-centrality
    """

    temp_df = pd.DataFrame()
    temp_df['Res. Name'] = input_class.mda_u.select_atoms(selection).residues.resnames
    temp_df['Res. Num.'] = input_class.mda_u.select_atoms(selection).residues.resnums
    temp_df['Acceptor Centrality'] = cent_acc
    temp_df['Donor Centrality'] = cent_don
    temp_df['Electrostatic Centrality'] = cent_don + cent_acc
    temp_df['CA Centrality'] = cent
    return temp_df







"""

MDiGest v.0.1.0

__version__ = 0.1.0
__author__ = Federica Maschietto <federica.maschietto@gmail.com>, Brandon Allen <bcallen95@gmail.com>

"""


from . import dim_reduction_plots
from . import general_plots
from . import plot_correlation

"""

----------

- dim_reduction plots   --> plots related to the dimensionality reduction module
- general_plots         --> plots
- plot_correlation      --> plots related to the correlation modules

"""



"""#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: fmaschietto, bcallen95"""

from mdigest.core.imports import  *

def log_progress(sequence, every=None, size=None, name='Items', userProgress=None):
    """
    Generates log progress bar

    See Also
    --------
    ``[**]`` this function was authored by Marcelo Melo as part of https://github.com/melomcr/dynetan

    """
    from ipywidgets import IntProgress, HTML, HBox, Label
    from IPython.display import display
    from numpy import mean as npmean
    from collections import deque
    from math import floor
    from datetime import datetime
    from string import Template

    is_iterator = False
    if size is None:
        try:
            size = len(sequence)
        except TypeError:
            is_iterator = True
    if size is not None:
        if every is None:
            if size <= 200:
                every = 1
            else:
                every = floor(float(size) * 0.005)  # every 0.5%, minimum is 1
    else:
        assert every is not None, 'sequence is iterator, set every'

    # For elapsed time
    initTime = datetime.now()
    totTime = "?"
    labTempl = Template(" (~ $min total time (min) ; $ell minutes elapsed)")

    # If provided, we use the objects already created.
    # If not provided, we create from scratch.
    if userProgress is None or userProgress == []:

        progress = IntProgress(min=0, max=1, value=1)

        label = HTML()
        labelTime = Label("")

        box = HBox(children=[label, progress, labelTime])

        if userProgress == []:
            userProgress.append(box)
        display(box)
    else:
        box = userProgress[0]

    if is_iterator:
        # progress = IntProgress(min=0, max=1, value=1)
        box.children[1].min = 0
        box.children[1].max = 1
        box.children[1].value = 1
        box.children[1].bar_style = 'info'
    else:
        # progress = IntProgress(min=0, max=size, value=0)
        box.children[1].min = 0
        box.children[1].max = size
        box.children[1].value = 0

        # For remaining time estimation
        deltas = deque()
        lastTime = None
        meandelta = 0

    index = 0
    try:
        for index, record in enumerate(sequence, 1):
            if index == 1 or index % every == 0:
                if is_iterator:
                    box.children[0].value = '{name}: {index} / ?'.format(
                        name=name,
                        index=index
                    )
                else:
                    box.children[1].value = index
                    box.children[0].value = u'{name}: {index} / {size}'.format(
                        name=name,
                        index=index,
                        size=size
                    )

                    # Estimates remaining time with average delta per iteration
                    # Uses (at most) the last 30 iterations
                    if len(deltas) == 101:
                        deltas.popleft()

                    if lastTime:
                        deltas.append((datetime.now() - lastTime).total_seconds())
                        meandelta = npmean(deltas) / 60.0  # From seconds to minute
                        totTime = round(meandelta * size / float(every), 3)  # Mean iteration for all iterations
                    else:
                        totTime = "?"  # First iteration has no time

                    lastTime = datetime.now()

                # All ellapsed time in minutes
                elapsed = round((datetime.now() - initTime).total_seconds() / 60.0, 3)

                box.children[2].value = labTempl.safe_substitute({"min": totTime,
                                                                  "ell": elapsed})

            yield record
    except:
        box.children[1].bar_style = 'danger'
        raise
    else:
        box.children[1].bar_style = 'success'
        box.children[1].value = index
        box.children[0].value = "{name}: {index}".format(
            name=name,
            index=str(index or '?')
        )


def dump(filepath, array_input):
    """
    Dump np.ndarray to file

    Parameters
    ----------
    filepath: str,
        output path

    array_input: np.ndarray,
        array to save

    Returns
    -------
    pickle binary output file
    """

    file_object = open(filepath, 'wb')
    pickle.dump(array_input, file_object)
    file_object.close()


def retrieve(filepath):
    """
    Retrive pickle object

    Parameters
    ----------
    filepath: str,
        path of file to read

    Returns
    -------
    :return content: np.ndarray,
        content
    """
    if os.path.exists(filepath) == 0:
        return None
    print("@>: loading pickle file...")
    file_object = open(filepath, 'rb')
    content = pickle.load(file_object)
    file_object.close()
    return content


def file_exists(filepath):
    """
    Check if file exists

    Parameters
    ----------
    filepath: str,
        path to file

    Returns
    -------
    :return file: bool,
        whether file is in path
    """
    from os.path import exists
    file = exists(filepath)
    return file


def folder_exists(path_to_folder):
    """
    Check if directory exists, create if not

    Parameters
    ----------
    path_to_folder: str,
        path to folder

    """
    from os import makedirs
    try:
        makedirs(path_to_folder)
    except OSError:
        # directory already exists
        pass


def keywithmaxval(d):
    """
    Create a list of the dict's keys and values; return the key with the max value
    """
    v = list(d.values())
    k = list(d.keys())
    return k[v.index(max(v))]


def get_or_minus1(f):
    """
    Assign to minus one if index is absent
    """
    try:
        return f()
    except IndexError:
        return -1


def normalize(arr):
    """
    Normalize array dividing by the sum.
    """
    sum = np.sum(arr)
    arr = arr / sum
    return arr


def normalize_array(vals):
    """
    Normalize array between -1 and 1.
    """
    maxim = np.max(vals)
    minim = np.min(vals)
    normalized = np.zeros(len(vals))
    for c in range(len(vals)):
        normalized[c] = 2 * ((vals[c] - minim) / (maxim - minim)) - 1
    return normalized


def intersection(lst1, lst2):
    """
    Find intersection between two lists
    """
    lst3 = [list(filter(lambda x: x in lst1, sublist)) for sublist in lst2]
    return lst3


def list2dict(listoflists):
    """
    Convert list to dictionary
    """
    dictoflists = {}
    for i, l in enumerate(listoflists):
        dictoflists[i] = l
    return dictoflists


def dict2list(dictoflists):
    """
    Convert dictionary to list
    """
    listoflists = []
    for k, v in dictoflists.items():
        listoflists.append(v)
    return listoflists


def partition2dict(partition):
    """
    Convert partitions to dictionary having nodes as keys and assigned community (partition)
    as values
    """
    tmp_ = {}
    for i, p in enumerate(partition):
        for node in p:
            tmp_.update({node: i})
    partition = tmp_
    return partition


def get_NGLselection_from_node(node_idx, atomsel, atom=True):
    """
    Create an atom selection (whole residue or single atom) for NGLView and an atom-selection object.
    """
    node = atomsel.atoms[node_idx]
    if atom:
        return " and ".join([str(node.resid), node.resname, "." + node.name])
    else:
        return " and ".join([str(node.resid), node.resname])


def get_selection_from_node(i, atomsel, atom=False):
    """
    Get the selection string from a node index: resname, resid, segid and name
    and return an atom-selection object.
    ``[**]`` function adapted from https://github.com/melomcr/dynetan
    """
    i = int(i)
    if i < 0:
        raise
    resname = atomsel.atoms[i].resname
    resid = str(atomsel.atoms[i].resid)
    segid = atomsel.atoms[i].segid
    atomname = atomsel.atoms[i].name

    if atom:
        return "resname " + resname + " and resid " + resid + " and segid " + segid + " and name " + atomname
    else:
        return "resname " + resname + " and resid " + resid + " and segid " + segid


def get_path(src, trg, selected_atomnodes, preds, rep=0):
    """
    Return an np.ndarray with the list of nodes that connect src (source) and trg (target).
    ``[**]`` function adapted from https://github.com/melomcr/dynetan
    """

    src = int(src)
    trg = int(trg)

    if src == trg:
        return []

    if src not in preds[rep].keys():
        return []

    if trg not in preds[rep][src].keys():
        return []

    if get_selection_from_node(src, selected_atomnodes) == get_selection_from_node(trg, selected_atomnodes):
        return []

    path = [trg]

    while path[-1] != src:
        path.append(preds[rep][src][path[-1]])

    return np.asarray(path)


"""#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: fmaschietto, bcallen95"""

from mdigest.core.imports import *
import mdigest.core.auxiliary as aux
from MDAnalysis.analysis.dihedrals import Janin
from MDAnalysis.analysis import dihedrals


class MDS_analysis:
    """
    Basic molecular dynamics analysis
    """

    def __init__(self):
        """
        Description
        -----------
        Given a molecular dynamics trajectory computes general anlalyses such as RMSF, RMSD, secondary structure analysis.
        """
        self.mds_data                  = None
        self.mda_u                     = None
        self.mdt_obj                   = None
        self.segIDs                    = None
        self.atm_str_sel               = None
        self.sys_str_sel               = None
        self.atom_group_selection      = None
        self.atom_type_selection       = None
        self.chi1_selection            = None

        # ---------------------------------------------#
        self.chi1_angles                        = None
        self.dihed_angles                       = None
        self.ss_stats                           = None
        self.rgyr_results                       = None
        self.rms_data                           = None
        self.nh_selections                      = None
        self.order_parameters                   = None
        self.pdb_df                             = None
        # ---------------------------------------------#

        self.natoms              = 0
        self.total_nframes       = 0
        self.nframes_per_replica = 0
        self.num_replicas        = 0
        self.nresidues           = 0
        self.initial             = 0
        self.final               = 0
        self.step                = 0
        self.window_span         = 0


    def set_num_replicas(self, num_replicas):
        self.num_replicas = num_replicas


    def load_system(self, topology, traj_files, inmem=True):
        self.mda_u = mda.Universe(topology,traj_files,in_memory=inmem)
        self.natoms = self.mda_u.atoms
        self.mdt_obj = md.load(traj_files, top = topology)
        return


    def align_traj(self, inmem=True, reference=None, selection='protein'):
        from MDAnalysis.analysis import align as mdaAlign
        # Set the first frame as reference for alignment
        self.mda_u.trajectory[0]
        if reference is None:
            alignment = mdaAlign.AlignTraj(self.mda_u, self.mda_u, select="segid " + " ".join(self.segIDs) + " and not (name H* or name [123]H*)",
                                       verbose=True, in_memory=inmem, weights="mass")
            alignment.run()
        else:
            alignment = mdaAlign.AlignTraj(self.mda_u, reference, select=selection,
                                       verbose=True, in_memory=inmem, weights="mass")
            alignment.run()
        return


    def set_node_type_sele(self, node_type_sele):
        self.atom_type_selection = node_type_sele
        return self.atom_type_selection


    def set_selection(self, atm_str_sel, sys_str_sel):
        self.atm_str_sel = atm_str_sel
        self.sys_str_sel = sys_str_sel
        return


    def get_universe(self):
        return self.mda_u


    def do_ss_calculation(self, simple=True):
        """
        Calculation of SS propensities using an MDTraj
        trajectory object. The `simple` option is a Boolean flag that allows for
        the choice of simple SS definitions (Helix, Strand, Coil) or the more
        descriptive SS definitions.

        Parameters
        ----------
        simple: bool
            If true, only 'H', 'E' and 'C' secondary structure assignments are used.

        Returns
        --------
        `self.ss_stats`: pd.DataFrame,
            SS assignments at each frame as well as the SS propensities computed over the course of the trajectory (i.e., % Helix, % Strand, % Coil).
        """
        def _vec_query(arr, my_dict):
            return np.vectorize(my_dict.__getitem__)(arr)

        mdt_u = self.mdt_obj
        sec_struct = md.compute_dssp(mdt_u, simplified=simple)
        if simple:
            ss_defs = {'H': 'Helix', 'E': 'Strand', 'C': 'Coil', 'NA': 'Not Protein'}
        else:
            ss_defs = {'H': 'Alpha Helix', 'B': 'Beta-Bridge',
                      'E': 'Extended Strand/Beta Ladder', 'G': '3/10 Helix',
                      'I': 'Pi Helix', 'T': 'Hydrogen-Bonded Turn',
                      'S': 'Bend', ' ': 'General Coil/Loops/Irregular Elements',
                      'NA': 'Not Protein'}
        ss_labels = _vec_query(sec_struct, ss_defs)
        nframe = sec_struct.shape[0]
        nres = sec_struct.shape[-1]
        resi = [i for i in mdt_u.top.residues]
        resname = [i.name for i in resi]
        resnum = [i.resSeq for i in resi]
        out_df = pd.DataFrame(data=ss_labels.T)
        out_df.insert(0, 'Res. Names', np.asarray(resname))
        out_df.insert(1, 'Res. Num.', np.asarray(resnum))
        out_df['% Helix'] = 0.00
        out_df['% Strand'] = 0.00
        out_df['% Coil'] = 0.00
        for mm in range(nres):
            temp_val = out_df.iloc[mm, 2:].value_counts()
            all_keys = list(temp_val.keys())
            for key in all_keys:
                if key == 'Coil':
                    out_df.loc[mm, '% Coil'] = temp_val[key]/nframe*100.00
                if key == 'Strand':
                    out_df.loc[mm, '% Strand'] = temp_val[key]/nframe*100.00
                if key == 'Helix':
                    out_df.loc[mm, '% Helix'] = temp_val[key]/nframe*100.00
        self.ss_stats = out_df


    def compute_radius_of_gyration(self, **kwargs):
        """
        Calculate the radius of gyration given an MDAnalysis.Universe and a dictionary of atom selections.
        This dictionary should take the following form:

        Parameters
        ----------
        kwargs: dict,
                - selection_id: string to describe the selection
                - selection_str: string used for creating selection with MDA.Universe
        """
        kw_args = kwargs['kwargs']
        temp_rgyr_dict = {}
        for ii, arg in enumerate(kw_args):
            temp_sel = self.mda_u.select_atoms(kw_args[arg])
            temp_list = []
            for ts in self.mda_u.trajectory:
                temp_list.append([self.mda_u.trajectory.time, temp_sel.radius_of_gyration()])
            temp_rgyr_dict[arg] = {'Selection String': kw_args[arg], 'rgyr': np.asarray(temp_list)}

        self.rgyr_results = temp_rgyr_dict
        return


    def calc_chi1(self):
        """
        Calculates the Chi 1 Angles for an MDAnalysis.Universe.
        Uses the following attributes:
        - self.mda_u: The MDAnalysis.Universe
        - self.chi1_selection: a string used to select the subset of the universe for which you want to calculate the Chi 1 Angles.

        Returns
        -------
        self.chi1_angles: an array representing the Chi 1 Angles of your selection at each frame of the trajectory.
        """

        temp_sel = self.mda_u.select_atoms(self.chi1_selection)
        temp_janin = Janin(temp_sel)
        janin_results = temp_janin.run()
        self.chi1_angles = janin_results.results['angles'][:, :, 0]
        return


    def calc_rms_quant(self, sel_str, **kwargs):
        """
        Calculate RMS Quantities (i.e., RMSD, RMSF).

        Parameters
        ----------
        sel_str: str or int,
            string/frame_index describing the reference with which to calculate the RMS quantity
        kwargs: dict
            example:
                ``kwargs={'Initial': 0, 'Average': 'average', 'Final': -1}``
            
        Returns
        -------
        self.rms_data --> A dict with the following structure:
            ``self.rms_data = {'RMSD': {selection_title: RMSD_Value}, 'RMSF': {selection_title: RMSF_Value}}``,
            where RMSD_Value/RMSF_Value are np.arrays containing the computed quantities.
        """

        # Extract the coordinates into array of shape (n_frames, n_atoms, 3)
        coords = aux._extract_coords(self.mda_u, sel_str)
        # Parse the key-word arguments:
        fancy_str = []
        reference_selections = []
        for ii, arg in enumerate(kwargs['kwargs']):
            fancy_str.append(arg)
            reference_selections.append(kwargs['kwargs'][arg])
        # Iterate through the reference_selection list:
        for ii, refsel in enumerate(reference_selections):
            results_dictionary = {}
            ref_coords = None
            if type(refsel) is not type(0):
                if (refsel.lower() == 'average') or (refsel == 'Average'):
                    ref_coords = np.mean(coords, axis=0)
            else:
                ref_coords = coords[refsel]
            n_atoms = len(ref_coords)
            n_frames = len(coords)
            ref_coords = ref_coords.reshape((1, n_atoms, 3))
            deviation = coords - ref_coords
            # Up to here, RMSD and RMSD are the same (both take norm, but different divisor)
            sq_fluct = np.linalg.norm(deviation, axis=-1)
            sq_dev = sq_fluct/n_atoms
            rmsd = np.sqrt(np.sum(sq_dev, axis=-1))
            mean_sq_fluct = np.sum(sq_fluct, axis=0)/n_frames
            rmsf = np.sqrt(mean_sq_fluct)
            results_dictionary['RMSD'] = {fancy_str[ii]: rmsd}
            results_dictionary['RMSF'] = {fancy_str[ii]: rmsf}

            if self.rms_data is None:
                self.rms_data = results_dictionary
            else:
                self.rms_data['RMSD'][fancy_str[ii]] = rmsd
                self.rms_data['RMSF'][fancy_str[ii]] = rmsf


    def do_dihedral_calcs(self, save_data=None):
        """
        Calculate Phi, Psi Angles from input:
        
        Parameters
        ----------
        save_data: dict or None,
            use for saving the output. Requires dict containing the keys:
            - "Directory": The desired directory to save the output to
            - "Output Descriptor": simple string to identify the system uniquely (for saving output)
        """
        str_sel = self.atm_str_sel
        temp_diheds = dihedrals.Ramachandran(self.mda_u.select_atoms(str_sel)).run()
        temp_phi = temp_diheds.angles[:, :, 0]
        temp_psi = temp_diheds.angles[:, :, 1]
        self.dihed_angles = {'Phi': temp_phi, 'Psi': temp_psi}
        if type(save_data) is None:
            pass
        else:
            temp_dir = save_data['Directory']
            if temp_dir.endswith('/'):
                temp_dir = temp_dir[:-1]
            else:
                temp_dir = temp_dir

            np.savez('{}/{}-Dihedrals.npz'.format(temp_dir, save_data['Output Descriptor']), phi=temp_phi, psi=temp_psi)


    def set_NH_selections(self, amide_string_list):
        """
            Expected input:
                - amide_string_list: a list of strings used to make the atom selection for your
                    backbone amide N atoms and the backbone amide H atoms
        """
        self.nh_selections = dict(zip(['N-Atom', 'H-Atom'], amide_string_list))
        return


    def calc_NH_order_params(self):
        """
        This function calculates the S2 amide order parameters according to
        [J. Am. Chem. Soc. 1998, 120, 5301-5311].
        Expected input:
        - set ``self.nh_selections`` before calling this function

        Returns
        -------
        self.order_parameters: pd.DataFrame

        TODO: This function will eventually need some kwargs to specify the N-atom/NH-atom selections.
        """


        def _calc_order_param(nh_vectors):
            """
            Function to actually calculate the order parameters from the
            amide N-H bond vectors.
            """
            oparam = 1.5 * (np.mean(nh_vectors[:, :, 0]**2, axis=0)**2 # <x^2>^2
                            + np.mean(nh_vectors[:, :, 1]**2, axis=0)**2 # <y^2>^2
                            + np.mean(nh_vectors[:, :, 2]**2, axis=0)**2 # <z^2>^2
                            + 2.*(np.mean(nh_vectors[:, :, 0]*nh_vectors[:, :, 1], axis=0)**2) # 2 <xy>^2
                            + 2.*(np.mean(nh_vectors[:, :, 0]*nh_vectors[:, :, 2], axis=0)**2) # 2 <xz>^2
                            + 2.*(np.mean(nh_vectors[:, :, 1]*nh_vectors[:, :, 2], axis=0)**2) # 2 <yz>^2
                            ) - 0.5
            return oparam

        # This function definitely can be improved!!!! As you can see, it currently
        # extracts the coordinates of all atoms and then grabs a subset using selection strings.
        if self.nh_selections is None:
            print('WARNING: You need to set your selection strings! Please call `self.set_NH_selections`')
        else:
            # Create atom selections for the backbone NH atoms
            n_idz = self.mda_u.select_atoms(self.nh_selections['N-Atom']).atoms.indices
            h_idz = self.mda_u.select_atoms(self.nh_selections['H-Atom']).atoms.indices
            if len(n_idz) == len(h_idz):
                pass
            else:
                n_idz = n_idz[1:]
            # Extract the coordinates from the universe.
            temp_xyz = aux._extract_coords(self.mda_u) # shape: (nframes, natoms, 3)
            # Use the atom indices from our atom selections to grab their respective coordinates.
            n_coords = temp_xyz[:, n_idz, :]
            h_coords = temp_xyz[:, h_idz, :]
            # Construct our Relative Vector pointing in the direction of the NH Bond: N ----> H
            nh_vects = h_coords - n_coords
            nh_norm = np.linalg.norm(nh_vects, axis=-1) # Need to make it a unit vector
            nh_vects[:, :, 0] /= nh_norm # Normalize the X-component
            nh_vects[:, :, 1] /= nh_norm # Normalize the Y-component
            nh_vects[:, :, 2] /= nh_norm # Normalize the Z-component
            temp_nh_o2 = _calc_order_param(nh_vects) # calculate S^2 with unit vectors.
            # NH Vector Order Parameter DF
            nh_ops = pd.DataFrame()
            nh_ops['Res. Num.'] = self.mda_u.atoms[n_idz].resnums
            nh_ops['Res. Name'] = self.mda_u.atoms[n_idz].resnames
            nh_ops['NH Order Param.'] = temp_nh_o2
            self.order_parameters = nh_ops
            return


    def stride_trajectory(self, initial=0, final=-1, step=1):
        """
        Stride trajectory

        Parameters
        ----------
        initial: int,
        final: int,
        step: int

        # TODO redundant with MDS, eventually remove and inherit mdigest.MDS object
        """

        # Select atom group of interest
        self.atom_group_selection = self.mda_u.select_atoms(self.atm_str_sel)
        if final < 0:
            final = self.mda_u.trajectory.n_frames
        else:
            final = final
        self.total_nframes = len(self.mda_u.trajectory)
        self.nframes_per_replica = len(self.mda_u.trajectory[initial:final:step])
        # Number or replicas
        num_replicas = self.num_replicas
        self.initial = initial
        self.final = final
        self.step =  step
        self.window_span = int(np.ceil(np.ceil((self.final - self.initial) / self.step)/ num_replicas))
        print('@>: number of frames:      %d' % self.total_nframes)
        print('@>: number of replicas:    %d' % self.num_replicas)
        print("@>: using window length of %d simulation steps" % self.window_span)
        print('@>: first frame:           %d' % initial)
        print('@>: last frame:            %d' % final  )
        print('@>: step:                  %d' % step   )
        self.nresidues = len(self.atom_group_selection.residues)
        print('@>: number of elements in selected atom group %d' % len(self.atom_group_selection))


    def _read_pdb_file(self, filename):
        """
        Utility Function for reading a reference PDB File into a Pandas DF
        For more complex pdb to file operations use mdigest.utils
        """
        out_lists = []
        with open(filename) as file:
            for line in file.readlines():
                if line.startswith('ATOM'):
                    out_lists.append([line[:6].strip(' '), int(line[6:11].strip(' ')),  line[12:16].strip(' '), line[16],
                                      line[17:20], line[21], int(line[22:26].strip(' ')),  line[26],
                                      [float(line[30:38]), float(line[38:46]), float(line[46:54])],
                                      float(line[54:60]), float(line[60:66]), line[72:76].strip(' '),
                                      line[76:78].strip(' '), [i if len(i) > 0 else '0' for i in line[78:80]][0]])
        column_names = ['Record Name', 'Atom Serial No.', 'Atom Name', 'Alt. Loc.', 'Res. Name', 'Chain ID',
                        'Res. No.', 'Ins. Code', 'Coordinates', 'Occ.', 'B Factor', 'Seg. ID', 'Atom Type', 'Charge']
        self.pdb_df = pd.DataFrame(out_lists, columns = column_names)
        return

    def _df_to_pdb(self, desired_outname):
        """
        Minimal function to write quantities like centralities, order_params into B-factor column
        For more complex pdb to file operations use mdigest.utils
        """
        df_vals = self.pdb_df.values
        with open(desired_outname, 'w') as file:
            for line in df_vals:
                strtline, serial, name, altloc, resname, chainid, reseq, icode, pos, occup, bfact, segid, atom_type, charge = line
                file.write('{:<6s}{:5d}  {:<4s}{:<1s}{:<3s}{:1s}{:4d}{:1s}   {:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}      {:<4s}{:<2s}{:<2s}\n'.format(strtline, serial, name, altloc, resname, chainid, reseq, icode, pos[0], pos[1], pos[2], occup, bfact, segid, atom_type, charge))
            file.write('TER   \n')
            file.write('END\n')
        return 'Your PDB File is written.'


    ######################## Load Class from cache ######################################
    def load_class_from_file(self, file_name_root):
        def _load_pickle_obj(name):
            with open(str(name)+'.pickle', 'rb') as f:
                return pickle.load(f)

        tmp_path = '{}_sel_str.txt'.format(file_name_root)
        if os.path.exists(tmp_path):
            with open(tmp_path, 'r') as f:
                self.atom_type_selection = ''.join(f.readlines()[0].split('_'))
        else:
            print('Atom Type Selection file does not exist.')

        tmp_chi1_path = '{}_chi1_sel_str.txt'.format(file_name_root)
        if os.path.exists(tmp_path):
            with open(tmp_chi1_path, 'r') as f:
                self.chi1_selection = ''.join(f.readlines()[0].split('_'))
        else:
            print('Chi1 Selection file does not exist.')

        tmp_order_param = '{}_NH_Amide_Order_Parameters.pickle.gz'.format(file_name_root)
        if os.path.exists(tmp_order_param):
            self.order_parameters = pd.read_pickle(tmp_order_param)
        else:
            print('File does not exist.')

        tmp_ss_path = '{}_SecStruct_Statistics.pickle.gz'.format(file_name_root)
        if os.path.exists(tmp_ss_path):
            self.ss_stats = pd.read_pickle(tmp_ss_path)
        else:
            print('Secondary Structure Data not found.')

        tmp_rgyr = '{}_radius_of_gyration'.format(file_name_root)
        if os.path.exists(tmp_rgyr+'.pickle'):
            self.rgyr_results = _load_pickle_obj(tmp_rgyr)
        else:
            print('Radius of Gyration data not found.')

        tmp_rms = '{}_rms_quantities'.format(file_name_root)
        if os.path.exists(tmp_rms+'.pickle'):
            self.rms_data = _load_pickle_obj(tmp_rms)
        else:
            print('RMS data not found.')

        with h5py.File(file_name_root + ".hf", "r") as f:
            print("@>: cached file found: loading ", file_name_root + '.hf')
            for key in f.keys():
                print("@>:", key, f[key].dtype, f[key].shape, f[key].size)

                if f[key].size > 1:
                    # Stores value in object
                    setattr(self, key, np.zeros(f[key].shape, dtype=f[key].dtype))

                    f[key].read_direct(getattr(self, key))
                else:
                    # For a *scalar* H5Py Dataset, we index using an empty souple.
                    setattr(self, key, f[key][()])

        if self.dihed_angles is not None:
            self.dihed_angles = {'Phi': self.dihed_angles['Phi'], 'Psi': self.dihed_angles['Psi']}

        return

    ######################## Save Class ######################################
    def save_class(self, file_name_root):

        def _save_pickle_obj(obj, name):
            with open('{}.pickle'.format(name), 'wb') as f:
                pickle.dump(obj, f)
            return

        if self.atom_type_selection is not None:
            with open(file_name_root + '_sel_str.txt', 'w+') as f:
                f.write('_'.join(self.atom_type_selection))
            f.close()

        if self.chi1_selection is not None:
            with open(file_name_root + '_chi1_sel_str.txt', 'w+') as f:
                f.write('_'.join(self.chi1_selection))
            f.close()

        if self.order_parameters is not None:
            self.order_parameters.to_pickle('{}_NH_Amide_Order_Parameters.pickle.gz'.format(file_name_root))

        if self.ss_stats is not None:
            self.ss_stats.to_pickle('{}_SecStruct_Statistics.pickle.gz'.format(file_name_root))

        if self.rgyr_results is not None:
            _save_pickle_obj(self.rgyr_results, '{}_radius_of_gyration'.format(file_name_root))

        if self.rms_data is not None:
            _save_pickle_obj(self.rms_data, '{}_rms_quantities'.format(file_name_root))

        # Opens the HDF5 file and store all data.
        with h5py.File(file_name_root + ".hf", "w") as f:
            if self.chi1_angles is not None:
                f_chi1_angles = f.create_dataset("chi1_angles",
                                                        shape=self.chi1_angles.shape,
                                                        dtype=self.chi1_angles.dtype,
                                                        data=self.chi1_angles)

            if self.dihed_angles is not None:
                f_phi_angles = f.create_dataset("phi_angles",
                                                        shape=self.dihed_angles['Phi'].shape,
                                                        dtype=self.dihed_angles['Phi'].dtype,
                                                        data=self.dihed_angles['Phi'])

                f_psi_angles = f.create_dataset("psi_angles",
                                                        shape=self.dihed_angles['Psi'].shape,
                                                        dtype=self.dihed_angles['Psi'].dtype,
                                                        data=self.dihed_angles['Psi'])



"""#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: fmaschietto, bcallen95"""

from mdigest.core.imports            import *
from MDAnalysis.analysis       import distances
from concurrent.futures.thread import ThreadPoolExecutor
import mdigest.core.savedata            as sd
import mdigest.core.toolkit              as tk
import mdigest.core.auxiliary  as aux


MAX_WORKERS = 4

class KS_Box:
    """
    Use as collector to free up memory
    """
    def __init__(self, KS, attrlist):
        import gc
        for attr in list(KS.__dict__().keys()):
            if attr in attrlist:
                setattr(self, attr, KS.__dict__()[attr])
        for v in dir(KS):
            if not v.startswith('__'):
                del v
        gc.collect()
        del KS



class KS_Energy:
    """General purpose class handling computation Kabsch-Sander Analysis over MD trajectories."""

    def __init__(self, MDSIM):
        """
        Description
        -----------
        General purpose class handling computation Kabsch-Sander Analysis over MD trajectories.

        Parameters
        ----------
        MDSIM: class object

        Attributes
        ----------
        self.indices: dict,
            maps a list of backbone atom label strings to a list of indices corresponding to the index of the selected
            backbone atom for each residue
        self.backbone_dictionary: dict,
            maps a list of backbone atom label strings to a list corresponding atom names (as found in the topology)
        self.is_protein: np.ndarray of dtype bool and shape (number of selected residues (protein or not))
            is_protein array is 1 if the corresponding residue has C,O,N,H protein backbone atoms in the topology, 0 otherwise
        self.is_proline: np.ndarray of dtype bool and shape (nresidues),
            is_protein array is 1 if the corresponding residue is a proline, 0 otherwise
        self.offset: int,
            default offset is 0, adjust if the resindices of the selcted atoms do not start at 0
        self.q1q2: np.ndarray of shape (nresidues, nresidues),
            entries ($q1q2_(ij)$)) are the products of the i-th CO and j-th NH residue charges extracted
            from the topology
        self.bb_distances_allrep: dict,
            stores distances between specified backbone atoms of each pair or residues used in KS
            calculation such as CN, ON, OH, CH distance matrices for each replica,
                - example: ``bb_distances_allrep['rep_0]['CN']`` returns a np.ndarray of shape (nresidues*nresidues) containing all pairwise C-N distances
        self.KS_energies_allrep: dict,
            stores KS energies for each replica of a given simulation object in the form of np.ndarray
            of shape (window_span, n_nresidues, nresidues),
                - example: access as KS_energies_allrep['rep_0] for KS energies of replica 0
        self.KS_DA_energies_allrep: dict,
            stores KS energies summed over columns(acceptor)/rows(donor), yielding acceptor/donor
            KS energy matrices of shape (window_span, n_nresidues)
                - example:
                    1) access as ``KS_DA_energies_allrep['rep_0']['donor']`` for donor energies of replica 0
                    2) access as ``KS_DA_energies_allrep['rep_0']['acceptor']`` for acceptor energies of replica 0
        self.KS_DA_LMI_corr_allrep: dict,
            stores per replica don_acc mutual information based generalized correlation matrices computed respectively from donor, acceptor and donor+acceptor KS energies
                - example:
                    access as ``KS_DA_MI_corr_allrep['rep_0']['don_acc']``
        self.KS_DA_LMI_corr_allrep: dict,
            stores per replica donor/acceptor/don_acc linearized mutual information based generalized correlation matrices computed respectively from donor, acceptor and donor+acceptor KS energies
                - example:
                    access as ``KS_DA_LMI_corr_allrep['rep_0']['donor']``
        self.KS_cov_allrep: dict
            stores per replica covariance matrix over donor, acceptor KS energy matrices computed along the trajectories
                - example:
                    access as ``KS_cov_allrep['rep_0']['donor']`` for the covariance of the donor KS energies ``KS_cov_allrep['rep_0']['acceptor']`` for the covariance of the acceptor KS energies
        self.eigvec_centrality_da_allrep: dict,
            stores the eigvector centrality array for each replica (rep) computed diagonalizing the ``KS_DA_LMI_corr_allrep[rep]['don_acc']`` matrix
        self.eigvec_centrality_don_allrep: dict,
            stores the eigvector centrality array for each replica (rep) computed diagonalizing the ``KS_DA_LMI_corr_allrep[rep]['donor']`` matrix
        self.eigvec_centrality_acc_allrep: dict,
            stores the eigvector centrality array for each replica (rep) computed diagonalizing the
            ``KS_DA_LMI_corr_allrep[rep]['acceptor']`` matrix
        self.eigvec_centrality_da_indep_sum_allrep: dict,
            stores the eigvector centrality array for each replica (rep) computed diagonalizing independently the ``KS_DA_LMI_corr_allrep[rep]['donor']`` and
            ``KS_DA_LMI_corr_allrep[rep]['acceptor']`` matrix and summing the 1st eigenvectors together

        Methods
        -------

        References
        ----------
        """
        self.mds_data                              = MDSIM.mds_data
        self.mda_u                                 = MDSIM.mda_u
        self.num_replicas                          = MDSIM.num_replicas
        self.initial                               = MDSIM.initial
        self.final                                 = MDSIM.final
        self.step                                  = MDSIM.step
        self.window_span                           = MDSIM.window_span
        self.atom_group_selstr                     = MDSIM.atom_group_selstr
        self.system_selstr                         = MDSIM.system_selstr
        self.nframes_per_replica                   = MDSIM.nframes_per_replica
        self.atom_group_selection                  = MDSIM.atom_group_selection
        self.nresidues                             = MDSIM.nresidues
        self.natoms                                = MDSIM.natoms
        self.total_nframes                         = MDSIM.total_nframes
        self.nodes_to_res_dictionary               = MDSIM.nodes_to_res_dictionary # useful for later plotting (match nodes id to resid)

        self.indices                               = None
        self.backbone_dictionary                   = None
        self.is_protein                            = None
        self.is_proline                            = None
        self.offset                                = 0
        self.q1q2                                  = None

        self.bb_distances_allrep                   = {}
        self.KS_energies_allrep                    = {}
        self.KS_DA_energies_allrep                 = {}
        self.KS_DA_MI_corr_allrep                  = {}
        self.KS_DA_LMI_corr_allrep                 = {}
        self.KS_cov_allrep                         = {}
        self.eigvec_centrality_da_allrep           = {}
        self.eigvec_centrality_don_allrep          = {}
        self.eigvec_centrality_acc_allrep          = {}
        self.eigvec_centrality_da_mean_allrep       = {}
        self.eigvec_centrality_da_indep_sum_allrep = {}


    def save_class(self, file_name_root='./output/cache/', save_space=False):
        """
        Save MDS class instances to file

        Parameters
        ----------
        file_name_root: str,
            path where to save class
        save_space: bool,
            if False bb_distances_allrep and KS_energies_allrep are not dumped to file
        """

        self.mds_data = sd.MDSdata()
        self.mds_data.is_protein                            = self.is_protein
        self.mds_data.is_proline                            = self.is_proline
        self.mds_data.q1q2                                  = self.q1q2
        self.mds_data.nodes_to_res_dictionary               = self.nodes_to_res_dictionary
        self.mds_data.bb_distances_allrep                   = self.bb_distances_allrep
        self.mds_data.KS_energies_allrep                    = self.KS_energies_allrep
        self.mds_data.KS_DA_energies_allrep                 = self.KS_DA_energies_allrep
        self.mds_data.KS_DA_MI_corr_allrep                  = self.KS_DA_MI_corr_allrep
        self.mds_data.KS_DA_LMI_corr_allrep                 = self.KS_DA_LMI_corr_allrep
        self.mds_data.KS_cov_allrep                         = self.KS_cov_allrep
        self.mds_data.eigvec_centrality_da_allrep           = self.eigvec_centrality_da_allrep
        self.mds_data.eigvec_centrality_don_allrep          = self.eigvec_centrality_don_allrep
        self.mds_data.eigvec_centrality_acc_allrep          = self.eigvec_centrality_acc_allrep
        self.mds_data.eigvec_centrality_da_mean_allrep       = self.eigvec_centrality_da_mean_allrep
        self.mds_data.eigvec_centrality_da_indep_sum_allrep = self.eigvec_centrality_da_indep_sum_allrep
        self.mds_data.save_to_file(file_name_root, save_space=save_space)


    def set_selection(self, atom_group_selstr, system_selstr='all'):
        """
        Set selection strings

        Parameters
        ----------
        system_selstr: str,
            selection string to be used for extracting a subset of the atoms (system) on which to perform analysis

        atom_group_selstr: str,
            selection string to be used for selecting a subset of atoms from the system
                - atom_str_sel: a list of four selection strings containing in order the N, O, C, H backbone selection strings, respectively.)

        Examples
        --------
        ``KS_Energy.set_selection(['protein and backbone and name N','protein and backbone and name O', 'protein and backbone and name C','protein and name H'], system_selstr='protein')``

        """
        keys = ['N-Backbone', 'O-Backbone', 'C-Backbone', 'H-Backbone']

        self.atom_group_selstr = dict(zip(keys, atom_group_selstr))
        self.system_selstr = system_selstr


    def set_charges_array(self, chargeOtimeschargeN):
        """
        Set charges array

        Parameters
        ----------
        chargeOtimeschargeN: np.ndarray,
            array of dimensions (nresidues, nresidues) entries (q1q2$_(ij)$)) are the products of the i-th CO and j-th NH residue charges extracted from the topology
        """
        self.q1q2 = chargeOtimeschargeN


    def set_offset(self, offset):
        """
        set offset

        Parameters
        ----------

        offset: int,
            set offset when the residue indexes in the topology start at number other than 0
            integer that should be subtracted to the first residue index have the resindices list start from 0
        """
        self.offset = offset


    def set_backbone_dictionary(self, backbone_dictionary):
        """
        Set backbone dictionary

        Parameters
        ----------

        backbone_dictionary: dict,
            backbone dictionary specifying the atom name of each backbone atom.
            Names should match those in the topology files

        Examples
        --------
        ``KS_energy.set_backbone_dictionary({'N-Backbone':'N', 'O-Backbone':'O','C-Backbone':'C', 'CA-Backbone':'CA', 'H-Backbone':'H'})``

        """
        self.backbone_dictionary = backbone_dictionary


    def _distances(self, indices1, indices2, coordinates, remap=False):
        """
        Compute distances
        """
        d = np.zeros((self.window_span, len(indices1), len(indices2)), dtype=float)

        for ts, coord in enumerate(coordinates):
            d[ts] = distances.distance_array(coord[indices1, 0:3], coord[indices2, 0:3], backend="OpenMP")
            if not remap:
                d[ts][:, np.where(self.is_protein == 0)[0]] = 0.
        return d

    def compute_distances_parallel(self, coordinates, remap=False):
        """
        Compute distances in parallel

        Parameters
        ----------
        beg: int,
            initial frame
        end: int,
            end frame
        stride: int,
            step
        remap: bool,
            if True assumes remapping is unneeded (all distance matrices have the same dimension)

        Returns
        -------
        bb_dist_dict: dict
            dictionary with CN, CH, OH, ON as keys and np.ndarrays with the corresponding distance arrays as values
        """

        print('@>: computing distances in parallel')

        C = np.asarray(self.indices['C-Backbone'], dtype=int)
        N = np.asarray(self.indices['N-Backbone'], dtype=int)
        O = np.asarray(self.indices['O-Backbone'], dtype=int)
        H = np.asarray(self.indices['H-Backbone'], dtype=int)


        func = self._distances

        results = {}
        with ThreadPoolExecutor(MAX_WORKERS) as executor:
            results['CN'] = executor.submit(func, C, N, coordinates, remap=True)
            results['CH'] = executor.submit(func, C, H, coordinates, remap=remap)
            results['OH'] = executor.submit(func, O, H, coordinates, remap=remap)
            results['ON'] = executor.submit(func, O, N, coordinates, remap=True)

        bb_dist_dict = {}

        bb_dist_dict.update({'CH': results['CH'].result()})
        bb_dist_dict.update({'ON': results['ON'].result()})
        bb_dist_dict.update({'OH': results['OH'].result()})
        bb_dist_dict.update({'CN': results['CN'].result()})
        return bb_dist_dict



    def compute_KS_energy(self, dist_dict, topology_charges=False):
        """
        Perform KS calculation

        Parameters
        ----------
        dist_dict: dict,
            single dictionary containing residue-to-residues backbone atom distances for a given replica
        topology_charges: bool,
            if True, ``self.q1q2`` is expected to be filled with charges array

        Returns
        -------
        KS_energies: np.ndarray,
            KS_energies
        """

        CH_dist = dist_dict['CH']
        ON_dist = dist_dict['ON']
        OH_dist = dist_dict['OH']
        CN_dist = dist_dict['CN']

        nframes = len(CN_dist)
        nresidues = len(CN_dist[0])
        KS_energies = np.zeros((nframes, nresidues, nresidues), dtype=np.float32)
        if not topology_charges:
            q1timesq2 = 0.42 * 0.20
        else:
            q1timesq2 = self.q1q2
        ffactor = 332
        for frame in trange(nframes):

            inv_ON = np.reciprocal(ON_dist[frame])
            inv_CH = np.reciprocal(CH_dist[frame])
            inv_OH = np.reciprocal(OH_dist[frame])
            inv_CN = np.reciprocal(CN_dist[frame])
            inv_OH = np.nan_to_num(inv_OH, nan=0.0, posinf=0.0, neginf=0.0)
            inv_CN = np.nan_to_num(inv_CN, nan=0.0, posinf=0.0, neginf=0.0)
            inv_CH = np.nan_to_num(inv_CH, nan=0.0, posinf=0.0, neginf=0.0)
            inv_ON = np.nan_to_num(inv_ON, nan=0.0, posinf=0.0, neginf=0.0)

            KS_energies[frame, :, :] = q1timesq2 * (inv_ON + inv_CH - inv_OH - inv_CN) * ffactor
        return KS_energies


    def prepare_kabsch_sander_arrays(self):
        """
        Prepare Kabsch-Sanders calculation
        """

        C_str  = self.backbone_dictionary['C-Backbone']
        CA_str = self.backbone_dictionary['CA-Backbone']
        N_str  = self.backbone_dictionary['N-Backbone']
        O_str  = self.backbone_dictionary['O-Backbone']
        H_str  = self.backbone_dictionary['H-Backbone']

        u = self.mda_u

        ca_indices, n_indices, c_indices, o_indices, h_indices, is_proline, is_protein = \
            [], [], [], [], [], [], []
        for residue in u.select_atoms(self.system_selstr).residues:


            ca = tk.get_or_minus1(lambda: [a.index for a in residue.atoms if a.name == CA_str][0])
            h  = tk.get_or_minus1(lambda: [a.index for a in residue.atoms if a.name == H_str][0])
            n  = tk.get_or_minus1(lambda: [a.index for a in residue.atoms if a.name == N_str][0])
            c  = tk.get_or_minus1(lambda: [a.index for a in residue.atoms if a.name == C_str][0])
            o  = tk.get_or_minus1(lambda: [a.index for a in residue.atoms if a.name == O_str][0])

            ca_indices.append(ca)
            n_indices.append(n)
            c_indices.append(c)
            o_indices.append(o)
            h_indices.append(h)

            is_proline.append(residue.resname == 'PRO')
            is_protein.append(h != -1 and n != -1 and c != -1 and o != -1)

        ca_indices  = np.array(ca_indices, np.int32)
        n_indices = np.array(n_indices, np.int32)
        c_indices = np.array(c_indices, np.int32)
        o_indices = np.array(o_indices, np.int32)
        h_indices = np.array(h_indices, np.int32)

        # ca_indices[ca_indices == -1] = ca_indices[ca_indices  == -1]
        n_indices[n_indices  == -1] = ca_indices[n_indices  == -1]
        c_indices[c_indices  == -1] = ca_indices[c_indices  == -1]
        o_indices[o_indices  == -1] = ca_indices[o_indices  == -1]
        h_indices[h_indices  == -1] = ca_indices[h_indices  == -1]

        is_proline = np.array(is_proline, np.int32)
        is_protein = np.array(is_protein, np.int32)

        self.indices = dict(zip(sorted(list(self.backbone_dictionary.keys())),
                                [c_indices, ca_indices, h_indices, n_indices, o_indices]))

        self.is_protein = is_protein
        self.is_proline = is_proline



    def compute_EEC(self, distance_matrix=None, loc_factor=None, don_acc=True):
        """
        Compute Electrostatic Eigenvector Centrality (EEC), donor/acceptor/don_acc centrality for each replica

        Parameters
        ----------

        distance_matrix: default None,
            provide distance matrix when loc_factor != 0 to zero out values
            adiacency matrix values corresponding to distances exceeding loc_factor
        loc_factor: float,
            filtering threshold for selection of specific correlation range
        don_acc: bool,
            whether to compute DA (donor_acceptor) and D+A (donor+acceptor) centralities
        """

        num_replicas = self.num_replicas
        ec_don_acc = None
        ec_don_acc_mean = None

        for win_idx in tk.log_progress(range(num_replicas), every=1, size=num_replicas, name="Window"):

            if loc_factor is not None:
                _, ec_donor = aux.compute_eigenvector_centrality(self.KS_DA_LMI_corr_allrep['rep_%d' %win_idx]['donor'], loc_factor=loc_factor,
                                                          distmat=distance_matrix)
                _, ec_acceptor = aux.compute_eigenvector_centrality(self.KS_DA_LMI_corr_allrep['rep_%d' %win_idx]['acceptor'], loc_factor=loc_factor,
                                                          distmat=distance_matrix)
                if don_acc:
                    _, ec_don_acc = aux.compute_eigenvector_centrality(self.KS_DA_LMI_corr_allrep['rep_%d' %win_idx]['don_acc'], loc_factor=loc_factor,
                                                          distmat=distance_matrix)
                    _, ec_don_acc_mean = aux.compute_eigenvector_centrality(self.KS_DA_LMI_corr_allrep['rep_%d' %win_idx]['mean_don_acc'], loc_factor=loc_factor,
                                                          distmat=distance_matrix)


            else:
                _, ec_donor = aux.compute_eigenvector_centrality(self.KS_DA_LMI_corr_allrep['rep_%d' %win_idx]['donor'], loc_factor=None,
                                                          distmat=None)
                _, ec_acceptor = aux.compute_eigenvector_centrality(self.KS_DA_LMI_corr_allrep['rep_%d' %win_idx]['acceptor'], loc_factor=None,
                                                          distmat=None)
                if don_acc:
                    _, ec_don_acc = aux.compute_eigenvector_centrality(self.KS_DA_LMI_corr_allrep['rep_%d' % win_idx]['don_acc'],
                                                          loc_factor=None, distmat=None)
                    _, ec_don_acc_mean = aux.compute_eigenvector_centrality(self.KS_DA_LMI_corr_allrep['rep_%d' %win_idx]['mean_don_acc'], loc_factor=loc_factor,
                                                          distmat=distance_matrix)

            ec_independent_sum = ec_donor + ec_acceptor
            # ec = np.array([*ec.values()], dtype=object)

            if don_acc:
                # from gcc joint gcc [don,acc]
                self.eigvec_centrality_da_allrep.update({'rep_%d' %win_idx: ec_don_acc})
                # from gcc gccdon + gccacc
                self.eigvec_centrality_da_mean_allrep.update({'rep_%d' % win_idx: ec_don_acc_mean})


            self.eigvec_centrality_don_allrep.update({'rep_%d' %win_idx: ec_donor})
            self.eigvec_centrality_acc_allrep.update({'rep_%d' %win_idx: ec_acceptor})
            self.eigvec_centrality_da_indep_sum_allrep.update({'rep_%d' %win_idx: ec_independent_sum})


    def KS_pipeline(self, topology_charges=False, covariance=False, MI=None, **kwargs):
        """
        KS Pipeline

        Parameters
        ----------
        topology_charges: bool,
            whether to use topology charges in KS calculation
        covariance: bool,
            whether to compute covariance of KS_energies
        MI: str or None,
            - if None skip computation of MI based correlation
            - if 'knn_arg1_arg2' compute MI using k=arg1, and estimator=arg2; default is 'knn_5_1'
        """

        def _compress_KS_energies(KS_E):
            """
            Compress KS_energies into donor and acceptor energies
            """
            KSenergy_acc = np.sum(KS_E, axis=1)
            KSenergy_don = np.sum(KS_E, axis=2)
            KS_energy_dict = {}
            KS_energy_dict.update({'donor': KSenergy_don,
                                   'acceptor': KSenergy_acc})
            return KS_energy_dict

        def _compute_KS_correlation(KS_energy_dict):
            """
            Compute Linearized mutual information based correlation of donor and acceptor KS energies using gaussian
            estimator to evaluate mutual information.
            """


            LMI_corr_donor = aux.compute_generalized_correlation_coefficients(KS_energy_dict['donor'],
                                                                              features_dimension=1, solver='gaussian',
                                                                              correction=False,
                                                                              subset=None)
            LMI_corr_acceptor = aux.compute_generalized_correlation_coefficients(KS_energy_dict['acceptor'],
                                                                              features_dimension=1, solver='gaussian',
                                                                              correction=False,
                                                                              subset=None)
            LMI_KS_corr = aux.compute_generalized_correlation_coefficients(KS_energy_dict['acceptor']+KS_energy_dict['donor'],
                                                                              features_dimension=1, solver='gaussian',
                                                                              correction=False,
                                                                              subset=None)


            LMI_KS_corr_sum = 0.5 * (LMI_corr_donor + LMI_corr_acceptor)
            LMI_KS_corr_dict = {}
            LMI_KS_corr_dict.update({'donor':    LMI_corr_donor,
                                     'acceptor': LMI_corr_acceptor,
                                     'don_acc':  LMI_KS_corr,
                                     'mean_don_acc':  LMI_KS_corr_sum})
            return LMI_KS_corr_dict

        def _compute_covar_KS_energies(KS_energy_dict):
            """
            Compute covariance of KS energies
            """
            cov_donor    = np.cov(KS_energy_dict['donor'],    rowvar=False, bias=True)
            cov_acceptor = np.cov(KS_energy_dict['acceptor'], rowvar=False, bias=True)
            # cov_don_acc  = np.cov(KS_energy_dict['don_acc'],  rowvar=False, bias=True)
            KS_cov_dict = {}
            KS_cov_dict.update({'donor': cov_donor})
            KS_cov_dict.update({'acceptor': cov_acceptor})
            # KS_cov_dict.update({'don_acc': cov_don_acc})
            return KS_cov_dict

        print('@>: prepare kabsch sanders calculation')
        self.prepare_kabsch_sander_arrays()
        print('@>: run KS calculation')


        for win_idx in tk.log_progress(range(self.num_replicas), every=1, size=self.num_replicas, name="Window"):
            beg = self.initial + self.window_span * win_idx
            end = (self.initial + self.window_span * (win_idx + 1))

            print("@>: KS energy calculation ...")
            print("@>: begin frame: %d" % beg)
            print("@>: end   frame: %d" % end)
            print("@>: step:        %d" % self.step)
            stride = self.step

            # precompute coordinates_of given replica:
            coordinates = np.zeros((self.window_span, self.mda_u.select_atoms('protein').atoms.n_atoms, 3), dtype=float)
            for ts in self.mda_u.trajectory[beg:end:stride]:
                coordinates[ts.frame] = ts.positions

            bb_distances_dict = self.compute_distances_parallel(coordinates)
            KS_energies = self.compute_KS_energy(bb_distances_dict, topology_charges=topology_charges)
            self.bb_distances_allrep.update({'rep_%d' % win_idx: bb_distances_dict})
            self.KS_energies_allrep.update({'rep_%d'  % win_idx: KS_energies})
            KS_DA_energies  = _compress_KS_energies(KS_energies)
            self.KS_DA_energies_allrep.update({'rep_%d' % win_idx: KS_DA_energies})
            KS_DA_LMI_corr = _compute_KS_correlation(KS_DA_energies)
            self.KS_DA_LMI_corr_allrep.update({'rep_%d' % win_idx: KS_DA_LMI_corr})

            # setup kwargs
            if MI:
                try:
                    correction = kwargs['correction']
                except KeyError:
                    correction = True

                try:
                    subset = kwargs['subset']
                except KeyError:
                    subset = None

                if MI is not None:
                    solver = MI
                else:
                    print('@>: default solver set to knn_5_1')
                    solver = 'knn_5_1'

                KS_DA_MI_corr = aux.compute_generalized_correlation_coefficients(
                    KS_DA_energies['acceptor'] + KS_DA_energies['donor'],
                    features_dimension=1, solver=solver, correction=correction, subset=subset)

                self.KS_DA_MI_corr_allrep.update({'rep_%d' % win_idx: {'don_acc': KS_DA_MI_corr}})
            if covariance:
                KS_cov = _compute_covar_KS_energies(KS_DA_energies)
                self.KS_cov_allrep.update({'rep_%d' % win_idx: KS_cov})




from setuptools import setup, find_packages

VERSION = '0.1.7'
DESCRIPTION = 'Best practices made easy for analysis of correlated motions from molecular dynamics simulations.'
LONG_DESCRIPTION = 'MDiGest is a best-practices-made-easy Python package that handles the most common issues in ' \
                   'the network-based analysis of correlated motions from molecular dynamics simulations.'

setup(
    name="mdigest",
    version=VERSION,
    author="Federica Maschietto, Brandon Allen",
    author_email="<federica.maschietto@gmail.com>, <brandon.allen@yale.edu>",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=['numpy', #<=1.24.3',
                      'numba', #>=0.56.4',
                      'scipy', #>= 1.10.1',
                      'python-louvain==0.15',
                      'pandas',
                      'seaborn',
                      'mdtraj',
                      'MDAnalysis',
                      'silx',
                      'nglview',
                      'networkx'],

    keywords=[
        'python',
        'correlation',
        'molecular dynamics',
        'MD trajectory analysis',
        'correlated motions',
        'network analysis',
        'community network'
    ],
    url="https://github.com/fmaschietto/MDiGest",
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
)



from mdigest.utils.imports import  *

def df_to_pdb(input_dataframe, desired_outname, protonation=False, charge_arr=False,
              split_by_chain=False, split_by_segid=False):
    """
    Advanced df to PDB with customization of chainIDs segIDs, and charges

    Parameters:
    input_dataframe: pd.DataFrame,
    desired_outname: str,
    protonation: bool,
    charge_arr: False or np.ndarray
    split_by_chain: bool,
    split_by_segid: bool
    """
    df_vals = input_dataframe.values

    fmt = {
        'ATOM': (
            "{record_name:6s}{serial:5d} {name:^3s}{altLoc:<1s}{resName:<4s}{chainID:1s}{resSeq:4d}{iCode:1s}   {posx:8.3f}{posy:8.3f}{posz:8.3f}{occupancy:6.2f}{tempFactor:6.2f}      {segID:<4s}{element:>2s}{charge:<2s}\n")}
    if (split_by_chain == False) and (split_by_segid == False):
        file = open(desired_outname + ".pdb", 'w')
        # print("no split")
    elif split_by_chain:
        chains = np.unique(input_dataframe['chainID']).astype(str)
        # print(chains)
        outfiles = {}
        for cID in chains:
            outfiles[cID] = [open(desired_outname + '_chain_%s.pdb' % cID, 'w')]
    elif split_by_segid:
        segids = np.unique(input_dataframe['segID']).astype(str)
        outfiles = {}
        for sID in segids:
            outfiles[sID] = [open(desired_outname + '_segid_%s.pdb' % sID, 'w')]

    for idx, line in enumerate(df_vals):
        record_name, serial, name, altloc, resname, chainid, reseq, icode, posx, posy, posz, occup, bfact, segid, atom_type, charge = line
        if charge_arr is not False:
            charge = str(charge_arr[idx])
        else:
            charge = charge

        vals = {'record_name': record_name, 'serial': serial}
        if len(name) >= 4:
            vals['name'] = name[:4]
        elif len(name) == 3:
            vals['name'] = ' {}'.format(name)
        elif len(name) == 2:
            vals['name'] = ' {} '.format(name)
        elif len(name) == 1:
            vals['name'] = ' {}  '.format(name)
        vals['altLoc'] = altloc
        vals['resName'] = resname
        vals['chainID'] = chainid
        vals['resSeq'] = reseq
        vals['iCode'] = icode
        vals['posx'] = posx
        vals['posy'] = posy
        vals['posz'] = posz  # don't take off atom so conversion works
        vals['occupancy'] = occup
        vals['tempFactor'] = bfact
        vals['segID'] = segid
        vals['element'] = atom_type
        vals['charge'] = '0'

        if vals['element'].strip(' ') == '':
            if (vals['resName'].strip(' ') == 'Cl' or vals['resName'].strip(' ') == 'CL' or vals['resName'].strip(
                    ' ') == 'Cl-'):
                vals['element'] = 'Cl'
            elif (vals['resName'].strip(' ') == 'Na' or vals['resName'].strip(' ') == 'NA' or vals['resName'].strip(
                    ' ') == 'Na+'):
                vals['element'] = 'Na'
            else:
                vals['element'] = vals['name'].strip(' ')[0]
                vals['charge'] = charge

        if protonation:
            if (vals['resName'].strip(' ') == 'ARG') and (vals['name'].strip(' ') == 'NH2'):
                vals['charge'] = '1+'
            elif (vals['resName'].strip(' ') == 'LYS') and (vals['name'].strip(' ') == 'NZ'):
                vals['charge'] = '1+'
            elif (vals['resName'].strip(' ') == 'HID') and (vals['name'].strip(' ') == 'ND1'):
                vals['charge'] = '1+'
            elif (vals['resName'].strip(' ') == 'HIE') and (vals['name'].strip(' ') == 'NE2'):
                vals['charge'] = '1+'
            elif (vals['resName'].strip(' ') == 'ASP') and (vals['name'].strip(' ') == 'OD2'):
                vals['charge'] = '1-'
            elif (vals['resName'].strip(' ') == 'GLU') and (vals['name'].strip(' ') == 'OE2'):
                vals['charge'] = '1-'
            else:
                vals['charge'] = '0'
                # .. _ATOM: http://www.wwpdb.org/documentation/file-format-content/format32/sect9.html#ATOM

        line = fmt['ATOM'].format(**vals)

        if (split_by_chain == False) and (split_by_segid == False):
            if len(line.strip(' ')) > 2:
                #print('@>',line)
                file.write(line)

        elif split_by_chain:
            if len(line.strip(' ')) > 2:
                chain = vals['chainID']
                if vals['chainID'] == chain:
                    outfiles[chain][0].write(line)
                chain = vals['chainID']

        elif split_by_segid:
            if len(line.strip(' ')) > 2:
                segid = vals['segID']
                if vals['segID'] == segid:
                    outfiles[segid][0].write(line)
                segid = vals['segID']

    if not ((split_by_chain == False) and (split_by_segid == False)):
        for k, v in outfiles.items():
            outfiles[k][0].close()
    else:
        file.close()

    return 'Your PDB File is written.'


def select_frame(mda_u, frame):
    u = mda_u.copy()
    return u.trajectory[frame]


def mdatopology_to_dataframe(mda_u, frame, selection='all'):
    df = pd.DataFrame()
    extract_element = lambda x: [e[0] for e in x]
    u = mda_u.copy()
    ag = u.select_atoms(selection)
    coordinates = select_frame(mda_u=u, frame=frame).positions
    try:
        record_types = ag.atoms.record_types
    except:
        record_types = ['ATOM'] * ag.atoms.n_atoms

    atom_serial  = np.arange(len(ag.atoms.ids))+1

    try:
        altloc   = ag.atoms.atoms.altLocs
    except:
        altloc   =  [' '] * ag.atoms.n_atoms
    resname      = ag.atoms.resnames
    try:
        icodes   = ag.atoms.icodes
    except:
        icodes   = [' '] * ag.atoms.n_atoms

    segids       = ag.atoms.segids
    try:
        chainIDs = ag.atoms.chainIDs
    except:
        chainIDs = [' '] * ag.atoms.n_atoms
    resids       = ag.atoms.resids
    coordx       = coordinates[:,0]
    coordy       = coordinates[:,1]
    coordz       = coordinates[:,2]

    try:
        occupancies  = ag.atoms.occupancies
    except:
        occupancies = np.zeros(ag.atoms.n_atoms, dtype=float)
    try:
        tempfactors  = ag.atoms.tempfactors
    except:
        tempfactors = np.zeros(ag.atoms.n_atoms, dtype=float)

    atomnames    = ag.atoms.names
    elements = extract_element(ag.atoms.names)

    df['record_names'] = record_types
    df['serial']  = atom_serial
    df['name'] = atomnames
    df['altLoc'] = altloc
    df['resName'] = resname
    df['chainID'] = chainIDs
    df['resSeq'] = resids
    df['iCode']   = icodes
    df['coordx']  = coordx
    df['coordy']  = coordy
    df['coordz']  = coordz
    df['occupancy']  = occupancies
    df['tempFactor'] = tempfactors
    df['segID']      = segids
    df['element'] = elements
    df['charge'] = np.zeros(len(resids), dtype=int)

    return df


def read_pdb_file(filename):
    out_lists = []
    with open(filename) as file:
        for line in file.readlines():
            if line.startswith('ATOM'):
                out_lists.append([line[:6].strip(' '), int(line[6:11].strip(' ')),  line[12:16].strip(' '), line[16], \
                                            line[17:20], line[21], int(line[22:26].strip(' ')),  line[26], \
                                            float(line[30:38]), float(line[38:46]), float(line[46:54]), \
                                             float(line[54:60]), float(line[60:66]), line[72:76].strip(' '), \
                                             line[76:78].strip(' '), [i if len(i) > 0 else '0' for i in line[78:80]][0]])
    column_names = ['Record Name', 'Atom Serial No.', 'Atom Name', 'Alt. Loc.', 'Res. Name', 'Chain ID',
                    'Res. No.', 'Ins. Code', 'posx', 'posy', 'posz', 'Occ.', 'B Factor', 'Seg. ID', 'Atom Type', 'Charge']
    out_data = pd.DataFrame(out_lists, columns = column_names)
    return out_data


def center_pdb(pdb_df, shift):
    pdb_df['posx'] = pdb_df['posx'].apply(lambda x: -x + shift[0] / 2)
    pdb_df['posy'] = pdb_df['posy'].apply(lambda x: -x + shift[0] / 2)
    pdb_df['posz'] = pdb_df['posz'].apply(lambda x: -x + shift[0] / 2)
    return pdb_df


def fill_topology_df(topo_df_mdt, coord_array):
    """
    This function completes the topology data frame from MDtraj for later
    saving the topology to PDB format.

    topo_df_mdt: topology from mdtraj.to_dataframe()[0]:
                      which only contains the following columns:
                      ['serial', 'name', 'element', 'resSeq', 'resName', 'chainID','segmentID']

    coord_array: array of coordinates from mdtraj.trajectory.xyz or average_structure.xyz

    Returns
    -------
    :return topo_df_mdt: topology dataframe completed of all fields
    """

    topo_df_mdt['record_name'] = ['ATOM']*len(topo_df_mdt)
    topo_df_mdt['altLoc'] = [' ']*len(topo_df_mdt)
    topo_df_mdt['posx'] = coord_array[:,0]
    topo_df_mdt['posy'] = coord_array[:,1]
    topo_df_mdt['posz'] = coord_array[:,2]

    # relpace chainID with alphabet letters in topo_df['chainID'] column
    topo_df_mdt['chainID'] = [chr(int(i)+65) for i in topo_df_mdt['chainID']]

    #rename colum of dataframe from segmentID to segID
    topo_df_mdt = topo_df_mdt.rename(columns={'segmentID':'segID'})

    # fill in the missing values of segID with chainID
    topo_df_mdt['segID'] = topo_df_mdt['chainID']
    topo_df_mdt['occupancy'] = [0.0]*len(topo_df_mdt)
    topo_df_mdt['tempFactor'] = [0.0]*len(topo_df_mdt)
    topo_df_mdt['iCode'] = [' ']*len(topo_df_mdt)
    topo_df_mdt['charge'] = [' ']*len(topo_df_mdt)

    # reorganize columns in topo_df dataframe according to PDB format
    col_names = ['record_name', 'serial', 'name', 'altLoc', 'resName', 'chainID', 'resSeq', 'iCode',
                 'posx', 'posy', 'posz', 'occupancy', 'tempFactor', 'segID', 'element', 'charge']

    # reorder columns in topo_df by names in column_names
    topo_df_mdt = topo_df_mdt[col_names]
    return topo_df_mdt

